# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import re
import tempfile
from pathlib import Path
from typing import Tuple, Optional, Union, List

import git
import yaml
import subprocess
from ogr.parsing import RepoUrl, parse_git_repo

from packit.constants import CENTOS_DOMAIN, CENTOS_STREAM_GITLAB
from packit.utils.commands import run_command

from packit.exceptions import PackitException

logger = logging.getLogger(__name__)


class RepositoryCache:
    """
    Cache for git repositories base on the reference option of `git clone`.

    * The cache is located in the specified directory
      and contains separate git repository for each project.
    * Project name is used to match the git project in the cache.
    """

    def __init__(self, cache_path: Path, add_new=False) -> None:
        self.cache_path = cache_path
        self.add_new = add_new

    @property
    def cached_projects(self) -> List[str]:
        """Project names we have in the cache."""
        if not self.cache_path.is_dir():
            self.cache_path.mkdir(parents=True)
        return [f.name for f in self.cache_path.iterdir() if f.is_dir()]

    def _clone(self, **kwargs) -> git.Repo:
        """Wrapper around git function so we are able to check the call in tests more easily."""
        return git.repo.Repo.clone_from(**kwargs)

    def get_repo(
        self,
        url: str,
        directory: Union[Path, str] = None,
    ) -> git.Repo:
        """
        Clone the repository.
        * If we have this repository in a cache, use the cached repo as a reference when cloning.
        * If we don't have this repository in a cache and {add_new} is True,
          clone the repository to cache first and then use it as a reference.

        :param url: will be used to clone the repo
        :param directory: target path for cloning the repository
        :return: cloned repository
        """
        directory = str(directory) if directory else tempfile.mkdtemp()

        if is_git_repo(directory=directory):
            logger.debug(f"Repo already exists in {directory}.")
            return git.repo.Repo(directory)

        logger.debug(f"Cloning repo {url} -> {directory}")

        project_name = RepoUrl.parse(url).repo
        reference_repo = self.cache_path.joinpath(project_name)
        if project_name not in self.cached_projects and self.add_new:
            logger.debug(f"Creating reference repo: {reference_repo}")
            self._clone(url=url, to_path=str(reference_repo), tags=True)

        if self.add_new or project_name in self.cached_projects:
            logger.debug(f"Using reference repo: {reference_repo}")
            return self._clone(
                url=url, to_path=directory, tags=True, reference=str(reference_repo)
            )

        return self._clone(url=url, to_path=directory, tags=True)


def is_git_repo(directory: Union[Path, str]) -> bool:
    """
    Test, if the directory is a git repo.
    (Has .git subdirectory?)
    """
    return Path(directory, ".git").is_dir()


def get_repo(url: str, directory: Union[Path, str] = None) -> git.Repo:
    """
    Use directory as a git repo or clone repo to the tempdir.
    """
    directory = str(directory) if directory else tempfile.mkdtemp()

    if is_git_repo(directory=directory):
        logger.debug(f"Repo already exists in {directory}.")
        repo = git.repo.Repo(directory)
    else:
        logger.debug(f"Cloning repo {url} -> {directory}")
        repo = git.repo.Repo.clone_from(url=url, to_path=directory, tags=True)

    return repo


def get_namespace_and_repo_name(url: str) -> Tuple[Optional[str], str]:
    parsed_git_repo = parse_git_repo(url)
    if parsed_git_repo is None or not parsed_git_repo.repo:
        raise PackitException(
            f"Invalid URL format, can't obtain namespace and repository name: {url}"
        )
    return parsed_git_repo.namespace, parsed_git_repo.repo


def is_a_git_ref(repo: git.Repo, ref: str) -> bool:
    try:
        commit = repo.commit(ref)
        return bool(commit)
    except git.BadName:
        return False


def get_default_branch(repository: git.Repo) -> str:
    """
    Returns default branch for newly created repos in the parent directory of
    passed in repository. Accepts `repository` to ensure the closest override of
    git configuration is used.

    Args:
        repository (git.Repo): Git repository closest to the directory where
            the configuration is applied.

    Returns:
        Default branch for new repos, if not supported or not configured returns
        `master`.
    """
    config = repository.config_reader()
    return config.get_value("init", "defaultBranch", "master")


def git_remote_url_to_https_url(inp: str) -> str:
    """
    turn provided git remote URL to https URL:
    returns empty string if the input can't be processed
    """
    logger.debug(f"Parsing git remote URL {inp!r} and converting it to https-like URL.")
    parsed_repo = parse_git_repo(inp)
    if not parsed_repo or not parsed_repo.hostname:
        logger.debug(f"{inp!r} is not an URL we recognize.")
        return ""

    if inp.startswith(("http", "https")):
        logger.debug(f"Provided input {inp!r} is an url.")
        return inp

    optional_suffix = ".git" if inp.endswith(".git") else ""
    url_str = "https://{}/{}/{}{}".format(
        parsed_repo.hostname, parsed_repo.namespace, parsed_repo.repo, optional_suffix
    )

    logger.debug(f"URL {inp!r} turned into HTTPS {url_str!r}")
    return url_str


def get_current_version_command(
    glob_pattern: str, refs: Optional[str] = "tags"
) -> List[str]:
    """
    Returns command that find latest git reference matching given pattern.

    :param glob_pattern: pattern that is used to find latest ref
    :param refs: specifies what kind of ref is used; \
        default is `"tags"` that searches through all tags (including non-annotated), \
        pass `None` to search only annotated tags or `"all"` to search through \
        all refs (including branches and remote refs)
    :return: command to find latest ref
    """
    return [
        "git",
        "describe",
        "--abbrev=0",
        f"--{refs}" if refs else "",
        "--match",
        glob_pattern,
    ]


def clone_centos_8_package(
    package_name: str,
    dist_git_path: Path,
    branch: str = "c8s",
    namespace: str = "rpms",
    stg: bool = False,
):
    """
    clone selected package from git.[stg.]centos.org
    """
    run_command(
        [
            "git",
            "clone",
            "-b",
            branch,
            f"https://git{'.stg' if stg else ''}.{CENTOS_DOMAIN}/{namespace}/{package_name}.git",
            str(dist_git_path),
        ]
    )


def clone_centos_9_package(
    package_name: str,
    dist_git_path: Path,
    branch: str = "c9s",
    namespace: str = "rpms",
    stg: bool = None,
):
    """
    clone selected package from git.[stg.]centos.org
    """
    if stg:
        logger.warning("There is no staging instance for CentOS Stream 9 dist-git.")
    run_command(
        [
            "git",
            "clone",
            "-b",
            branch,
            f"https://{CENTOS_STREAM_GITLAB}/{namespace}/{package_name}.git",
            str(dist_git_path),
        ]
    )


def clone_fedora_package(
    package_name: str,
    dist_git_path: Path,
    branch: str = None,
    namespace: str = "rpms",
    stg: bool = False,
):
    """
    clone selected package from Fedora's src.fedoraproject.org
    """
    command = [
        "git",
        "clone",
        f"https://src{'.stg' if stg else ''}.fedoraproject.org/{namespace}/{package_name}.git",
        str(dist_git_path),
    ]
    if branch:
        command += ["-b", branch]

    run_command(command)


def create_new_repo(cwd: Path, switches: List[str]):
    subprocess.check_call(["git", "init"] + switches + [str(cwd)])
    # TODO: Replace with -b / --initial-branch in `git init` when possible
    if "--bare" not in switches:
        subprocess.check_call(["git", "checkout", "-b", "main"], cwd=cwd)
    else:
        subprocess.check_call(
            ["git", "symbolic-ref", "HEAD", "refs/heads/main"], cwd=cwd
        )


def git_patch_ish(patch: str) -> str:
    """
    Massage patch to look like a Git-style patch, so that it can
    be passed to 'git patch-id' in order to calculate a patch-id.

    :param patch: Patch to transform.
    :return: Transformed patch.
    """
    # Prettend as if format is 'diff --git'
    pattern = re.compile(r"^diff -\w+ ", flags=re.MULTILINE)
    repl = r"diff --git "
    patch = re.sub(pattern, repl, patch)

    # Remove timestamps from comparison lines
    pattern = re.compile(r"^((---|\+\+\+) .+)\t\d{4}.+$", flags=re.MULTILINE)
    repl = r"\1"
    patch = re.sub(pattern, repl, patch)

    # Add missing 'diff --git' lines
    if "diff --git " not in patch:
        # Timestamps (see above) already need to be removed
        # for this substitution pattern to work.
        pattern = re.compile(r"(\n--- (.+)\n\+\+\+ (.+)\n)")
        repl = r"\ndiff --git \2 \3\1"
        patch = re.sub(pattern, repl, patch)

    return patch


def get_message_from_metadata(metadata: dict, header: Optional[str] = None) -> str:
    if not isinstance(metadata, dict):
        raise PackitException(
            f"We can save only dictionaries to metadata. Not {metadata}"
        )

    content = (
        yaml.dump(metadata, indent=4, default_flow_style=False) if metadata else ""
    )
    if not header:
        return content

    return f"{header}\n\n{content}"


def get_metadata_from_message(commit: git.Commit) -> Optional[dict]:
    """
    Tries to load yaml format from the git message.

    We are skipping first line until
    the rest of the content is yaml-loaded to dictionary (yaml object type).

    If nothing found, we return None.

    Reference:
    https://gitpython.readthedocs.io/en/stable/reference.html
    ?highlight=archive#module-git.objects.commit

    e.g.:

    I)
    key: value
    another: value
    -> {"key": "value", "another": "value"}

    II)
    On sentence.

    key: value
    another: value
    -> {"key": "value", "another": "value"}

    III)
    A lot of
    text

    before keys.

    key: value
    another: value
    -> {"key": "value", "another": "value"}

    IV)
    Other values are supported as well:

    key:
    - first
    - second
    - third

    :param commit: git.Commit object
    :return: dict loaded from message if it satisfies the rules above
    """
    splitted_message = commit.message.split("\n")

    for i in range(len(splitted_message)):
        message_part = "\n".join(splitted_message[i:])
        try:
            loaded_part = yaml.safe_load(message_part)
        except yaml.YAMLError:
            continue

        if isinstance(loaded_part, dict):
            return loaded_part

    return None



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import json
import logging
from pathlib import Path
from typing import Optional, List, Dict, Union, Set

from ogr.abstract import GitProject
from yaml import safe_load

from packit.actions import ActionName
from packit.config.common_package_config import CommonPackageConfig
from packit.config.job_config import JobConfig, get_default_jobs, JobType
from packit.config.notifications import NotificationsConfig
from packit.config.sources import SourcesItem
from packit.sync import SyncFilesItem
from packit.constants import CONFIG_FILE_NAMES
from packit.exceptions import PackitConfigException

logger = logging.getLogger(__name__)


class PackageConfig(CommonPackageConfig):
    """
    Config class for upstream/downstream packages;
    this is the config people put in their repos
    """

    def __init__(
        self,
        config_file_path: Optional[str] = None,
        specfile_path: Optional[str] = None,
        synced_files: Optional[List[SyncFilesItem]] = None,
        jobs: Optional[List[JobConfig]] = None,
        dist_git_namespace: str = None,
        upstream_project_url: str = None,  # can be URL or path
        upstream_package_name: str = None,
        downstream_project_url: str = None,
        downstream_package_name: str = None,
        dist_git_base_url: str = None,
        create_tarball_command: List[str] = None,
        current_version_command: List[str] = None,
        actions: Dict[ActionName, Union[str, List[str]]] = None,
        upstream_ref: Optional[str] = None,
        allowed_gpg_keys: Optional[List[str]] = None,
        create_pr: bool = True,
        sync_changelog: bool = False,
        spec_source_id: str = "Source0",
        upstream_tag_template: str = "{version}",
        archive_root_dir_template: str = "{upstream_pkg_name}-{version}",
        patch_generation_ignore_paths: List[str] = None,
        notifications: Optional[NotificationsConfig] = None,
        copy_upstream_release_description: bool = False,
        sources: Optional[List[SourcesItem]] = None,
    ):
        super().__init__(
            config_file_path=config_file_path,
            specfile_path=specfile_path,
            synced_files=synced_files,
            dist_git_namespace=dist_git_namespace,
            upstream_project_url=upstream_project_url,
            upstream_package_name=upstream_package_name,
            downstream_project_url=downstream_project_url,
            downstream_package_name=downstream_package_name,
            dist_git_base_url=dist_git_base_url,
            create_tarball_command=create_tarball_command,
            current_version_command=current_version_command,
            actions=actions,
            upstream_ref=upstream_ref,
            allowed_gpg_keys=allowed_gpg_keys,
            create_pr=create_pr,
            sync_changelog=sync_changelog,
            spec_source_id=spec_source_id,
            upstream_tag_template=upstream_tag_template,
            archive_root_dir_template=archive_root_dir_template,
            patch_generation_ignore_paths=patch_generation_ignore_paths,
            notifications=notifications,
            copy_upstream_release_description=copy_upstream_release_description,
            sources=sources,
        )
        self.jobs: List[JobConfig] = jobs or []

        if current_version_command:
            logger.warning(
                "current_version_command is deprecated and will be removed. Please "
                "switch to the `get-current-version` action: https://packit.dev/docs/actions/"
            )

        if create_tarball_command:
            logger.warning(
                "create_tarball_command is deprecated and will be removed. Please"
                "switch to the `create-archive` action: https://packit.dev/docs/actions/"
            )

    def __repr__(self):
        return (
            "PackageConfig("
            f"config_file_path='{self.config_file_path}', "
            f"specfile_path='{self.specfile_path}', "
            f"synced_files='{self.synced_files}', "
            f"jobs='{self.jobs}', "
            f"dist_git_namespace='{self.dist_git_namespace}', "
            f"upstream_project_url='{self.upstream_project_url}', "
            f"upstream_package_name='{self.upstream_package_name}', "
            f"downstream_project_url='{self.downstream_project_url}', "
            f"downstream_package_name='{self.downstream_package_name}', "
            f"dist_git_base_url='{self.dist_git_base_url}', "
            f"create_tarball_command='{self.create_tarball_command}', "
            f"current_version_command='{self.current_version_command}', "
            f"actions='{self.actions}', "
            f"upstream_ref='{self.upstream_ref}', "
            f"allowed_gpg_keys='{self.allowed_gpg_keys}', "
            f"create_pr='{self.create_pr}', "
            f"sync_changelog='{self.sync_changelog}', "
            f"spec_source_id='{self.spec_source_id}', "
            f"upstream_tag_template='{self.upstream_tag_template}', "
            f"archive_root_dir_template={self.archive_root_dir_template}', "
            f"patch_generation_ignore_paths='{self.patch_generation_ignore_paths}', "
            f"copy_upstream_release_description='{self.copy_upstream_release_description}',"
            f"sources='{self.sources}')"
        )

    @classmethod
    def get_from_dict(
        cls,
        raw_dict: dict,
        config_file_path: str = None,
        repo_name: str = None,
        spec_file_path: str = None,
    ) -> "PackageConfig":
        # required to avoid cyclical imports
        from packit.schema import PackageConfigSchema

        if config_file_path and not raw_dict.get("config_file_path", None):
            raw_dict.update(config_file_path=config_file_path)

        # we need to process defaults first so they get propagated to JobConfigs

        if "jobs" not in raw_dict:
            # we want default jobs to go through the proper parsing process
            raw_dict["jobs"] = get_default_jobs()

        if not raw_dict.get("specfile_path", None):
            if spec_file_path:
                raw_dict["specfile_path"] = spec_file_path
            else:
                raise PackitConfigException("Spec file was not found!")

        if not raw_dict.get("upstream_package_name", None) and repo_name:
            raw_dict["upstream_package_name"] = repo_name

        if not raw_dict.get("downstream_package_name", None) and repo_name:
            raw_dict["downstream_package_name"] = repo_name

        return PackageConfigSchema().load(raw_dict)

    def get_copr_build_project_value(self) -> Optional[str]:
        """
        get copr project name from this first copr job
        this is only used when invoking copr builds from CLI
        """
        projects_list = [
            job.metadata.project
            for job in self.jobs
            if job.type == JobType.copr_build and job.metadata.project
        ]
        if not projects_list:
            return None

        if len(set(projects_list)) > 1:
            logger.warning(
                f"You have defined multiple copr projects to build in, we are going "
                f"to pick the first one: {projects_list[0]}, reorder the job definitions"
                f" if this is not the one you want."
            )
        return projects_list[0]

    def get_propose_downstream_dg_branches_value(self) -> Optional[Set]:
        for job in self.jobs:
            if job.type == JobType.propose_downstream:
                return job.metadata.dist_git_branches
        return set()

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        logger.debug(f"our configuration:\n{self.__dict__}")
        logger.debug(f"the other configuration:\n{other.__dict__}")
        return (
            self.specfile_path == other.specfile_path
            and self.synced_files == other.synced_files
            and self.jobs == other.jobs
            and self.dist_git_namespace == other.dist_git_namespace
            and self.upstream_project_url == other.upstream_project_url
            and self.upstream_package_name == other.upstream_package_name
            and self.downstream_project_url == other.downstream_project_url
            and self.downstream_package_name == other.downstream_package_name
            and self.dist_git_base_url == other.dist_git_base_url
            and self.current_version_command == other.current_version_command
            and self.create_tarball_command == other.create_tarball_command
            and self.actions == other.actions
            and self.allowed_gpg_keys == other.allowed_gpg_keys
            and self.create_pr == other.create_pr
            and self.sync_changelog == other.sync_changelog
            and self.spec_source_id == other.spec_source_id
            and self.upstream_tag_template == other.upstream_tag_template
            and self.copy_upstream_release_description
            == other.copy_upstream_release_description
            and self.sources == other.sources
        )


def find_packit_yaml(
    *directory,
    try_local_dir_first=False,
    try_local_dir_last=False,
) -> Path:
    """
    find packit.yaml in provided directories: if a file matches, it's picked
    if no config is found, raise PackitConfigException

    :param directory: a list of dirs where we should find
    :param try_local_dir_first: try in cwd first
    :param try_local_dir_last: try cwd last
    :return: Path to the config
    """
    directories = [Path(config_dir) for config_dir in directory]
    cwd = Path.cwd()

    if try_local_dir_first and try_local_dir_last:
        logger.error(
            "Ambiguous usage of 'try_local_dir_first' and 'try_local_dir_last'."
        )

    if try_local_dir_first:
        if cwd in directories:
            directories.remove(cwd)
        directories.insert(0, cwd)

    if try_local_dir_last:
        if cwd in directories:
            directories.remove(cwd)
        directories.append(cwd)

    for config_dir in directories:
        for config_file_name in CONFIG_FILE_NAMES:
            config_file_name_full = config_dir / config_file_name
            if config_file_name_full.is_file():
                logger.debug(f"Local package config found: {config_file_name_full}")
                return config_file_name_full
    raise PackitConfigException("No packit config found.")


def load_packit_yaml(config_file_path: Path) -> Dict:
    """
    load provided packit.yaml, raise PackitConfigException if something is wrong

    :return: Dict with the file content
    """
    try:
        return safe_load(config_file_path.read_text())
    except Exception as ex:
        logger.error(f"Cannot load package config {config_file_path}.")
        raise PackitConfigException(f"Cannot load package config: {ex!r}.")


def get_local_package_config(
    *directory,
    repo_name: str = None,
    try_local_dir_first=False,
    try_local_dir_last=False,
    package_config_path: str = None,
) -> PackageConfig:
    """
    find packit.yaml in provided dirs, load it and return PackageConfig

    :param directory: a list of dirs where we should find
    :param repo_name: name of the git repository (default for project name)
    :param try_local_dir_first: try in cwd first
    :param try_local_dir_last: try cwd last
    :param package_config_path: Path to package configuration file
    :return: local PackageConfig
    """

    if package_config_path:
        config_file_name = Path(package_config_path)
    else:
        config_file_name = find_packit_yaml(
            *directory,
            try_local_dir_first=try_local_dir_first,
            try_local_dir_last=try_local_dir_last,
        )

    loaded_config = load_packit_yaml(config_file_name)
    return parse_loaded_config(
        loaded_config=loaded_config,
        config_file_path=config_file_name.name,
        repo_name=repo_name,
        spec_file_path=str(get_local_specfile_path(config_file_name.parent)),
    )


def get_package_config_from_repo(
    project: GitProject, ref: str = None, spec_file_path: Optional[str] = None
) -> Optional[PackageConfig]:
    for config_file_name in CONFIG_FILE_NAMES:
        try:
            config_file_content = project.get_file_content(
                path=config_file_name, ref=ref
            )
        except FileNotFoundError:
            # do nothing
            pass
        else:
            logger.debug(
                f"Found a config file {config_file_name!r} "
                f"on ref {ref!r} "
                f"of the {project.full_repo_name!r} repository."
            )
            break
    else:
        logger.warning(
            f"No config file ({CONFIG_FILE_NAMES}) found on ref {ref!r} "
            f"of the {project.full_repo_name!r} repository."
        )
        return None

    try:
        loaded_config = safe_load(config_file_content)
    except Exception as ex:
        logger.error(f"Cannot load package config {config_file_name!r}. {ex}")
        raise PackitConfigException(
            f"Cannot load package config {config_file_name!r}. {ex}"
        )
    if not spec_file_path:
        spec_file_path = get_specfile_path_from_repo(project=project, ref=ref)

    return parse_loaded_config(
        loaded_config=loaded_config,
        config_file_path=config_file_name,
        repo_name=project.repo,
        spec_file_path=spec_file_path,
    )


def parse_loaded_config(
    loaded_config: dict,
    config_file_path: Optional[str] = None,
    repo_name: Optional[str] = None,
    spec_file_path: Optional[str] = None,
) -> PackageConfig:
    """Tries to parse the config to PackageConfig."""
    logger.debug(f"Package config:\n{json.dumps(loaded_config, indent=4)}")

    try:
        return PackageConfig.get_from_dict(
            raw_dict=loaded_config,
            config_file_path=config_file_path,
            repo_name=repo_name,
            spec_file_path=spec_file_path,
        )
    except Exception as ex:
        logger.error(f"Cannot parse package config. {ex}.")
        raise PackitConfigException(f"Cannot parse package config: {ex!r}.")


def get_local_specfile_path(dir: Path, exclude: List[str] = None) -> Optional[Path]:
    """
    Get the path (relative to dir) of the local spec file if present.
    If the spec is not found in dir directly, try to search it recursively (rglob)
    :param dir: to find the spec file in
    :param exclude: don't include files found in these dirs (default "tests")
    :return: path (relative to dir) of the first found spec file
    """
    files = [path.relative_to(dir) for path in dir.glob("*.spec")] or [
        path.relative_to(dir) for path in dir.rglob("*.spec")
    ]

    if len(files) > 0:
        # Don't take files found in exclude
        sexclude = set(exclude) if exclude else {"tests"}
        files = [f for f in files if f.parts[0] not in sexclude]

        logger.debug(f"Local spec files found: {files}. Taking: {files[0]}")
        return files[0]

    return None


def get_specfile_path_from_repo(project: GitProject, ref: str = None) -> Optional[str]:
    """
    Get the path of the spec file in the given repo if present.
    :param project: GitProject
    :param ref: git ref (defaults to repo's default branch)
    :return: str path of the spec file or None
    """
    spec_files = project.get_files(ref=ref, filter_regex=r".+\.spec$")

    if not spec_files:
        logger.debug(f"No spec file found in {project.full_repo_name!r}")
        return None
    return spec_files[0]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock

from packit.config import Config, PackageConfig
from packit.distgit import DistGit
from packit.local_project import LocalProject


@pytest.mark.parametrize(
    "title, description, branch, prs, exists",
    [
        (
            "Update",
            "Upstream tag: 0.4.0\nUpstream commit: 6957453b",
            "f31",
            [
                flexmock(
                    title="Update",
                    target_branch="f31",
                    description="Upstream tag: 0.4.0\nUpstream commit: 6957453b",
                )
            ],
            True,
        ),
        (
            "Update",
            "Upstream tag: 0.4.0\nUpstream commit: 6957453b",
            "f31",
            [
                flexmock(
                    title="Update",
                    target_branch="f31",
                    description="Upstream tag: 0.4.0\nUpstream commit: 8957453b",
                )
            ],
            False,
        ),
        (
            "Update",
            "Upstream tag: 0.4.0\nUpstream commit: 6957453b",
            "f32",
            [
                flexmock(
                    title="Update",
                    target_branch="f31",
                    description="Upstream tag: 0.4.0\nUpstream commit: 6957453b",
                )
            ],
            False,
        ),
    ],
)
def test_pr_exists(title, description, branch, prs, exists):
    local_project = LocalProject(
        git_project=flexmock(service="something", get_pr_list=lambda: prs),
        refresh=False,
    )
    distgit = DistGit(
        config=flexmock(Config()),
        package_config=flexmock(PackageConfig()),
        local_project=local_project,
    )
    assert distgit.pr_exists(title, description, branch) == exists



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from pathlib import Path, PosixPath
from typing import Optional

import pytest
from flexmock import flexmock
from marshmallow import ValidationError

from ogr.abstract import GitProject, GitService
from packit.actions import ActionName
from packit.config import (
    JobType,
    JobConfigTriggerType,
    JobConfig,
    get_package_config_from_repo,
)
from packit.config.job_config import JobMetadataConfig
from packit.config.package_config import (
    get_specfile_path_from_repo,
    PackageConfig,
    get_local_specfile_path,
    get_local_package_config,
)
import packit.config.package_config
from packit.config.sources import SourcesItem
from packit.schema import PackageConfigSchema
from packit.sync import SyncFilesItem
from tests.spellbook import UP_OSBUILD, SYNC_FILES
from tests.unit.test_config import (
    get_job_config_dict_full,
    get_job_config_dict_simple,
    get_job_config_simple,
    get_job_config_full,
    get_default_job_config,
    get_job_config_dict_build_for_branch,
    get_job_config_build_for_branch,
)


@pytest.fixture()
def job_config_simple():
    return get_job_config_simple()


@pytest.mark.parametrize(
    "files,expected",
    [(["foo.spec"], "foo.spec"), ([], None)],
)
def test_get_specfile_path_from_repo(files, expected):
    gp = flexmock(GitProject)
    gp.should_receive("full_repo_name").and_return("a/b")
    gp.should_receive("get_files").and_return(files)
    git_project = GitProject(repo="", service=GitService(), namespace="")
    assert get_specfile_path_from_repo(project=git_project) == expected


@pytest.mark.parametrize(
    "package_config, project",
    [
        (
            PackageConfig(
                specfile_path="xxx",
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.pull_request,
                    )
                ],
            ),
            None,
        ),
        (
            PackageConfig(
                specfile_path="xxx",
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.pull_request,
                        metadata=JobMetadataConfig(project="example"),
                    )
                ],
            ),
            "example",
        ),
        (
            PackageConfig(
                specfile_path="xxx",
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.release,
                        metadata=JobMetadataConfig(project="example1"),
                    ),
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.pull_request,
                        metadata=JobMetadataConfig(project="example2"),
                    ),
                ],
            ),
            "example1",
        ),
    ],
)
def test_project_from_copr_build_job(package_config, project):
    config_project_value = package_config.get_copr_build_project_value()
    assert config_project_value == project


@pytest.mark.parametrize(
    "package_config, expected",
    [
        (
            PackageConfig(
                specfile_path="xxx",
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        metadata=JobMetadataConfig(
                            dist_git_branches=[
                                "example",
                            ]
                        ),
                    )
                ],
            ),
            {
                "example",
            },
        ),
        (
            PackageConfig(
                specfile_path="xxx",
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        metadata=JobMetadataConfig(
                            dist_git_branches=[
                                "example1",
                                "example2",
                            ]
                        ),
                    )
                ],
            ),
            {"example1", "example2"},
        ),
        (
            PackageConfig(
                specfile_path="xxx",
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                    )
                ],
            ),
            set(),
        ),
    ],
)
def test_dg_branches_from_propose_downstream_job(package_config, expected):
    branches = package_config.get_propose_downstream_dg_branches_value()
    assert branches == expected


def test_package_config_equal(job_config_simple):
    assert PackageConfig(
        specfile_path="fedora/package.spec",
        synced_files=[SyncFilesItem(src=["packit.yaml"], dest="packit.yaml")],
        jobs=[job_config_simple],
        downstream_package_name="package",
        create_pr=True,
    ) == PackageConfig(
        specfile_path="fedora/package.spec",
        synced_files=[SyncFilesItem(src=["packit.yaml"], dest="packit.yaml")],
        jobs=[job_config_simple],
        downstream_package_name="package",
        create_pr=True,
    )


@pytest.mark.parametrize(
    "not_equal_package_config",
    [
        PackageConfig(
            specfile_path="fedora/other-package.spec",
            synced_files=[
                SyncFilesItem(src=["a"], dest="a"),
                SyncFilesItem(src=["b"], dest="b"),
            ],
            jobs=[get_job_config_simple()],
        ),
        PackageConfig(
            specfile_path="fedora/package.spec",
            synced_files=[SyncFilesItem(src=["c"], dest="c")],
            jobs=[get_job_config_simple()],
        ),
        PackageConfig(
            specfile_path="fedora/package.spec",
            synced_files=[
                SyncFilesItem(src=["a"], dest="a"),
                SyncFilesItem(src=["b"], dest="b"),
            ],
            jobs=[get_job_config_full()],
        ),
        PackageConfig(
            specfile_path="fedora/package.spec",
            synced_files=[
                SyncFilesItem(src=["c"], dest="c"),
                SyncFilesItem(src=["d"], dest="d"),
            ],
            jobs=[get_job_config_full()],
            create_pr=False,
        ),
    ],
)
def test_package_config_not_equal(not_equal_package_config):
    config = PackageConfig(
        specfile_path="fedora/package.spec",
        synced_files=[
            SyncFilesItem(src=["c"], dest="c"),
            SyncFilesItem(src=["d"], dest="d"),
        ],
        jobs=[get_job_config_full()],
        create_pr=True,
    )
    assert config != not_equal_package_config


@pytest.mark.parametrize(
    "raw,is_valid",
    [
        (
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": "fedora/foobar.spec",
            },
            False,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/foobar.spec"],
            },
            True,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/foobar.spec", "somefile", "somedirectory"],
                "jobs": [],
            },
            True,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "metadata": {"dist-git-branch": "fedora-all"},
                    }
                ],
            },
            True,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "metadata": {"dist_git_branches": ["fedora-all", "epel-8"]},
                    }
                ],
            },
            True,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "metadata": {"targets": "fedora-stable"},
                    }
                ],
            },
            True,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "metadata": {
                            "targets": ["fedora-stable", "fedora-development"]
                        },
                    }
                ],
            },
            True,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "metadata": {
                            "targets": ["f31", "f32"],
                            "timeout": 123,
                            "owner": "santa",
                            "project": "gifts",
                        },
                    }
                ],
            },
            True,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/foobar.spec"],
                "actions": {
                    "pre-sync": "some/pre-sync/command --option",
                    "get-current-version": "get-me-version",
                },
            },
            True,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/foobar.spec"],
                "actions": {
                    "pre-sync": "some/pre-sync/command --option",
                    "unknown-action": "nothing",
                },
            },
            False,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "actions": ["actions" "has", "to", "be", "key", "value"],
                "jobs": [{"job": "asd", "trigger": "qwe"}],
            },
            False,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "notifications": {"pull_request": {"successful_build": False}},
            },
            True,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "notifications": {"pull_request": {"successful_build": "nie"}},
            },
            False,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "notifications": {"pull_request": False},
            },
            False,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "sources": [{"path": "example_path"}],
            },
            False,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "sources": [{"url": "example_url"}],
            },
            False,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "sources": {"path": "example_path", "url": "example_url"},
            },
            False,
        ),
    ],
)
def test_package_config_validate(raw, is_valid):
    if not is_valid:
        with pytest.raises((ValidationError, ValueError)):
            PackageConfig.get_from_dict(raw)
    else:
        PackageConfig.get_from_dict(raw)


@pytest.mark.parametrize(
    "raw,is_valid",
    [
        (
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "unknown": "key",
                    }
                ],
            },
            False,
        ),
        (
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "metadata": {"unknown": "key"},
                    }
                ],
            },
            False,
        ),
    ],
)
def test_package_config_validate_unknown_key(raw, is_valid):
    if not is_valid:
        with pytest.raises((ValidationError, ValueError)):
            PackageConfig.get_from_dict(raw)
    else:
        PackageConfig.get_from_dict(raw)


@pytest.mark.parametrize(
    "raw",
    [
        # {"specfile_path": "test/spec/file/path", "something": "different"},
        {
            "specfile_path": "test/spec/file/path",
            "jobs": [{"trigger": "release", "release_to": ["f28"]}],
        }
    ],
)
def test_package_config_parse_error(raw):
    with pytest.raises(Exception):
        PackageConfig.get_from_dict(raw_dict=raw)


@pytest.mark.parametrize(
    "raw,expected",
    [
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec"],
                "jobs": [get_job_config_dict_full()],
                "downstream_package_name": "package",
                "create_pr": False,
            },
            PackageConfig(
                specfile_path="fedora/package.spec",
                downstream_package_name="package",
                create_pr=False,
                synced_files=[
                    SyncFilesItem(
                        src=["fedora/package.spec"], dest="fedora/package.spec"
                    )
                ],
                jobs=[
                    get_job_config_full(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        create_pr=False,
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            )
                        ],
                    )
                ],
            ),
            id="specfile_path+synced_files+job_config_full+downstream_package_name+create_pr",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": [
                    "fedora/package.spec",
                    "somefile",
                    "other",
                    "directory/files",
                ],
                "jobs": [get_job_config_dict_simple()],
                "downstream_package_name": "package",
            },
            PackageConfig(
                downstream_package_name="package",
                specfile_path="fedora/package.spec",
                synced_files=[
                    SyncFilesItem(
                        src=["fedora/package.spec"], dest="fedora/package.spec"
                    ),
                    SyncFilesItem(src=["somefile"], dest="somefile"),
                    SyncFilesItem(src=["other"], dest="other"),
                    SyncFilesItem(src=["directory/files"], dest="directory/files"),
                ],
                jobs=[
                    get_job_config_simple(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                            SyncFilesItem(src=["somefile"], dest="somefile"),
                            SyncFilesItem(src=["other"], dest="other"),
                            SyncFilesItem(
                                src=["directory/files"], dest="directory/files"
                            ),
                        ],
                    )
                ],
            ),
            id="specfile_path+synced_files+job_config_dict_simple+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec"],
                "jobs": [get_job_config_dict_full()],
                "downstream_package_name": "package",
            },
            PackageConfig(
                downstream_package_name="package",
                specfile_path="fedora/package.spec",
                synced_files=[
                    SyncFilesItem(
                        src=["fedora/package.spec"], dest="fedora/package.spec"
                    )
                ],
                jobs=[
                    get_job_config_full(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            )
                        ],
                    )
                ],
            ),
            id="specfile_path+synced_files(spec_only)+job_config_full+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec", "somefile"],
                "jobs": [get_job_config_dict_full()],
                "downstream_package_name": "package",
            },
            PackageConfig(
                downstream_package_name="package",
                specfile_path="fedora/package.spec",
                synced_files=[
                    SyncFilesItem(
                        src=["fedora/package.spec"], dest="fedora/package.spec"
                    ),
                    SyncFilesItem(src=["somefile"], dest="somefile"),
                ],
                jobs=[
                    get_job_config_full(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                            SyncFilesItem(src=["somefile"], dest="somefile"),
                        ],
                    )
                ],
            ),
            id="specfile_path+synced_files+job_config_full+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec"],
                "jobs": [get_job_config_dict_full()],
                "upstream_project_url": "https://github.com/asd/qwe",
                "upstream_package_name": "qwe",
                "dist_git_base_url": "https://something.wicked",
                "downstream_package_name": "package",
            },
            PackageConfig(
                downstream_package_name="package",
                specfile_path="fedora/package.spec",
                upstream_project_url="https://github.com/asd/qwe",
                upstream_package_name="qwe",
                dist_git_base_url="https://something.wicked",
                synced_files=[
                    SyncFilesItem(
                        src=["fedora/package.spec"], dest="fedora/package.spec"
                    )
                ],
                jobs=[
                    get_job_config_full(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        upstream_project_url="https://github.com/asd/qwe",
                        upstream_package_name="qwe",
                        dist_git_base_url="https://something.wicked",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            )
                        ],
                    )
                ],
            ),
            id="specfile_path+synced_files+job_config_dict_full+upstream_project_url"
            "+upstream_package_name+dist_git_base_url+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "actions": {
                    "pre-sync": "some/pre-sync/command --option",
                    "get-current-version": "get-me-version",
                },
                "jobs": [],
                "upstream_project_url": "https://github.com/asd/qwe",
                "upstream_package_name": "qwe",
                "dist_git_base_url": "https://something.wicked",
                "downstream_package_name": "package",
            },
            PackageConfig(
                specfile_path="fedora/package.spec",
                actions={
                    ActionName.pre_sync: "some/pre-sync/command --option",
                    ActionName.get_current_version: "get-me-version",
                },
                jobs=[],
                upstream_project_url="https://github.com/asd/qwe",
                upstream_package_name="qwe",
                dist_git_base_url="https://something.wicked",
                downstream_package_name="package",
            ),
            id="specfile_path+actions+empty_jobs+upstream_project_url"
            "+upstream_package_name+dist_git_base_url+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec"],
                "downstream_package_name": "package",
            },
            PackageConfig(
                downstream_package_name="package",
                specfile_path="fedora/package.spec",
                synced_files=[
                    SyncFilesItem(
                        src=["fedora/package.spec"], dest="fedora/package.spec"
                    )
                ],
                jobs=get_default_job_config(
                    downstream_package_name="package",
                    specfile_path="fedora/package.spec",
                    synced_files=[
                        SyncFilesItem(
                            src=["fedora/package.spec"], dest="fedora/package.spec"
                        )
                    ],
                ),
            ),
            id="specfile_path+synced_files+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "spec_source_id": 3,
                "jobs": [get_job_config_dict_build_for_branch()],
            },
            PackageConfig(
                specfile_path="fedora/package.spec",
                spec_source_id="Source3",
                jobs=[
                    get_job_config_build_for_branch(
                        specfile_path="fedora/package.spec",
                        spec_source_id="Source3",
                    )
                ],
            ),
            id="specfile_path+get_job_config_dict_build_for_branch",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "sync_changelog": True,
                "jobs": [get_job_config_dict_simple()],
            },
            PackageConfig(
                specfile_path="fedora/package.spec",
                sync_changelog=True,
                jobs=[
                    get_job_config_simple(
                        specfile_path="fedora/package.spec", sync_changelog=True
                    )
                ],
            ),
            id="sync_changelog_true",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [get_job_config_dict_simple()],
            },
            PackageConfig(
                specfile_path="fedora/package.spec",
                sync_changelog=False,
                jobs=[
                    get_job_config_simple(
                        specfile_path="fedora/package.spec", sync_changelog=False
                    )
                ],
            ),
            id="sync_changelog_false_by_default",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "sources": [
                    {
                        "path": "rsync-3.1.3.tar.gz",
                        "url": "https://git.centos.org/sources/rsync/c8s/82e7829",
                    }
                ],
                "jobs": [],
            },
            PackageConfig(
                specfile_path="fedora/package.spec",
                sources=[
                    SourcesItem(
                        path="rsync-3.1.3.tar.gz",
                        url="https://git.centos.org/sources/rsync/c8s/82e7829",
                    ),
                ],
                jobs=[],
            ),
            id="sources",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "metadata": {
                            "fmf_url": "https://example.com",
                            "fmf_ref": "test_ref",
                        },
                    },
                ],
            },
            PackageConfig(
                specfile_path="fedora/package.spec",
                sync_changelog=False,
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        specfile_path="fedora/package.spec",
                        trigger=JobConfigTriggerType.release,
                        metadata=JobMetadataConfig(
                            fmf_url="https://example.com", fmf_ref="test_ref"
                        ),
                    ),
                ],
            ),
            id="sync_changelog_false_by_default",
        ),
    ],
)
def test_package_config_parse(raw, expected):
    package_config = PackageConfig.get_from_dict(raw_dict=raw)
    assert package_config
    # tests for https://github.com/packit/packit-service/pull/342
    if expected.jobs:
        for j in package_config.jobs:
            assert j.type
    assert package_config == expected


@pytest.mark.parametrize(
    "raw,expected",
    [
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "build",
                        "trigger": "release",
                        "specfile_path": "somewhere/package.spec",
                    }
                ],
            },
            PackageConfig(
                specfile_path="fedora/package.spec",
                jobs=[
                    JobConfig(
                        type=JobType.build,
                        trigger=JobConfigTriggerType.release,
                        specfile_path="somewhere/package.spec",
                    )
                ],
            ),
            id="override-specfile_path",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["x", "y"],
                "actions": {"post-upstream-clone": "ls"},
                "spec_source_id": "Source0",
                "jobs": [
                    {
                        "job": "build",
                        "trigger": "release",
                        "specfile_path": "somewhere/package.spec",
                        "synced_files": ["a", "b", "c"],
                        "actions": {"create-archive": "ls"},
                        "spec_source_id": "Source1",
                    }
                ],
            },
            PackageConfig(
                specfile_path="fedora/package.spec",
                synced_files=[SyncFilesItem([x], x) for x in ("x", "y")],
                actions={ActionName.post_upstream_clone: "ls"},
                jobs=[
                    JobConfig(
                        type=JobType.build,
                        trigger=JobConfigTriggerType.release,
                        specfile_path="somewhere/package.spec",
                        synced_files=[SyncFilesItem([x], x) for x in ("a", "b", "c")],
                        actions={ActionName.create_archive: "ls"},
                        spec_source_id="Source1",
                    )
                ],
            ),
            id="override-alot",
        ),
    ],
)
def test_package_config_overrides(raw, expected):
    package_config = PackageConfig.get_from_dict(raw_dict=raw)
    assert package_config == expected


@pytest.mark.parametrize(
    "raw,err_message",
    [
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [{"job": "build", "trigger": "release", "actions": ["a"]}],
            },
            "'dict' required, got <class 'list'>.",
            id="bad_actions",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "jobs": [{"job": "build", "trigger": "release", "synced_files": "a"}],
            },
            "ValidationError",
            id="bad_synced_files",
        ),
    ],
)
def test_package_config_overrides_bad(raw, err_message):
    with pytest.raises(ValidationError) as ex:
        PackageConfig.get_from_dict(raw_dict=raw)
    assert err_message in str(ex)


@pytest.mark.parametrize(
    "raw,expected",
    [
        (
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec"],
                "jobs": [],
            },
            PackageConfig(
                specfile_path="fedora/package.spec",
                synced_files=[
                    SyncFilesItem(
                        src=["fedora/package.spec"], dest="fedora/package.spec"
                    )
                ],
                downstream_package_name="package",
                upstream_package_name="package",
            ),
        )
    ],
)
def test_package_config_upstream_and_downstream_package_names(raw, expected):
    package_config = PackageConfig.get_from_dict(raw_dict=raw, repo_name="package")
    assert package_config
    assert package_config == expected


def test_dist_git_package_url():
    di = {
        "dist_git_base_url": "https://packit.dev/",
        "downstream_package_name": "packit",
        "dist_git_namespace": "awesome",
        "synced_files": ["fedora/foobar.spec"],
        "specfile_path": "fedora/package.spec",
        "create_pr": False,
    }
    new_pc = PackageConfig.get_from_dict(di)
    pc = PackageConfig(
        dist_git_base_url="https://packit.dev/",
        downstream_package_name="packit",
        dist_git_namespace="awesome",
        synced_files=[
            SyncFilesItem(src=["fedora/foobar.spec"], dest="fedora/foobar.spec")
        ],
        specfile_path="fedora/package.spec",
        create_pr=False,
        jobs=get_default_job_config(
            dist_git_base_url="https://packit.dev/",
            downstream_package_name="packit",
            dist_git_namespace="awesome",
            synced_files=[
                SyncFilesItem(src=["fedora/foobar.spec"], dest="fedora/foobar.spec")
            ],
            specfile_path="fedora/package.spec",
            create_pr=False,
        ),
    )
    assert new_pc.specfile_path.endswith("fedora/package.spec")
    assert pc.specfile_path.endswith("fedora/package.spec")
    assert pc == new_pc
    assert pc.dist_git_package_url == "https://packit.dev/awesome/packit.git"
    assert new_pc.dist_git_package_url == "https://packit.dev/awesome/packit.git"
    assert not pc.create_pr


@pytest.mark.parametrize(
    "content,project,mock_spec_search,spec_path_option,spec_path",
    [
        (
            "synced_files:\n"
            "  - packit.spec\n"
            "  - src: .packit.yaml\n"
            "    dest: .packit2.yaml",
            GitProject(repo="", service=GitService(), namespace=""),
            True,
            None,
            "packit.spec",
        ),
        (
            "synced_files:\n"
            "  - packit.spec\n"
            "  - src: .packit.yaml\n"
            "    dest: .packit2.yaml",
            GitProject(repo="", service=GitService(), namespace=""),
            False,
            "packit.spec",
            "packit.spec",
        ),
    ],
)
def test_get_package_config_from_repo(
    content,
    project: GitProject,
    mock_spec_search: bool,
    spec_path: Optional[str],
    spec_path_option: Optional[str],
):
    gp = flexmock(GitProject)
    gp.should_receive("full_repo_name").and_return("a/b")
    gp.should_receive("get_file_content").and_return(content)
    if mock_spec_search:
        gp.should_receive("get_files").and_return(["packit.spec"]).once()
    config = get_package_config_from_repo(
        project=project, spec_file_path=spec_path_option
    )
    assert isinstance(config, PackageConfig)
    assert config.specfile_path == spec_path
    assert config.synced_files == [
        SyncFilesItem(src=["packit.spec"], dest="packit.spec"),
        SyncFilesItem(src=[".packit.yaml"], dest=".packit2.yaml"),
    ]
    assert config.create_pr


@pytest.mark.parametrize(
    "content",
    [
        "{}",
        "{jobs: [{job: build, trigger: commit}]}",
        "{downstream_package_name: horkyze, jobs: [{job: build, trigger: commit}]}",
        "{upstream_package_name: slize, jobs: [{job: build, trigger: commit}]}",
    ],
)
def test_get_package_config_from_repo_spec_file_not_defined(content):
    specfile_path = "packit.spec"
    gp = flexmock(GitProject)
    gp.should_receive("full_repo_name").and_return("a/b")
    gp.should_receive("get_file_content").and_return(content)
    gp.should_receive("get_files").and_return([specfile_path])
    git_project = GitProject(repo="", service=GitService(), namespace="")
    config = get_package_config_from_repo(project=git_project)
    assert isinstance(config, PackageConfig)
    assert Path(config.specfile_path).name == specfile_path
    assert config.create_pr
    for j in config.jobs:
        assert j.specfile_path == specfile_path
        assert j.downstream_package_name == config.downstream_package_name
        assert j.upstream_package_name == config.upstream_package_name


@pytest.mark.parametrize(
    "package_config, all_synced_files",
    [
        (
            PackageConfig(
                config_file_path="packit.yaml",
                specfile_path="file.spec",
                synced_files=[SyncFilesItem(src=["file.spec"], dest="file.spec")],
            ),
            [
                SyncFilesItem(src=["file.spec"], dest="file.spec"),
                SyncFilesItem(src=["packit.yaml"], dest="packit.yaml"),
            ],
        ),
        (
            PackageConfig(
                config_file_path="packit.yaml",
                specfile_path="file.spec",
                downstream_package_name="package",
                synced_files=[SyncFilesItem(src=["file.spec"], dest="package.spec")],
            ),
            [
                SyncFilesItem(src=["file.spec"], dest="package.spec"),
                SyncFilesItem(src=["packit.yaml"], dest="packit.yaml"),
            ],
        ),
        (
            PackageConfig(
                config_file_path="packit.yaml",
                specfile_path="file.spec",
                downstream_package_name="package",
                synced_files=[SyncFilesItem(src=["file.txt"], dest="file.txt")],
            ),
            [
                SyncFilesItem(src=["file.txt"], dest="file.txt"),
                SyncFilesItem(src=["file.spec"], dest="package.spec"),
                SyncFilesItem(src=["packit.yaml"], dest="packit.yaml"),
            ],
        ),
        (
            PackageConfig(
                config_file_path="packit.yaml",
                specfile_path="file.spec",
                synced_files=[],
                downstream_package_name="package",
            ),
            [
                SyncFilesItem(src=["file.spec"], dest="package.spec"),
                SyncFilesItem(src=["packit.yaml"], dest="packit.yaml"),
            ],
        ),
    ],
)
def test_get_all_files_to_sync(package_config, all_synced_files):
    assert package_config.get_all_files_to_sync() == all_synced_files


def test_notifications_section():
    pc = PackageConfig.get_from_dict({"specfile_path": "package.spec"})
    assert not pc.notifications.pull_request.successful_build


def test_get_local_specfile_path():
    assert str(get_local_specfile_path(UP_OSBUILD)) == "osbuild.spec"
    assert not get_local_specfile_path(SYNC_FILES)


@pytest.mark.parametrize(
    "directory, local_first,local_last,config_file_name,res_pc_path",
    [
        ([], False, True, None, Path.cwd() / ".packit.yaml"),
        ([], False, False, "different_conf.yaml", "different_conf.yaml"),
    ],
)
def test_get_local_package_config_path(
    directory, local_first, local_last, config_file_name, res_pc_path
):

    flexmock(PosixPath).should_receive("is_file").and_return(True)

    (
        flexmock(packit.config.package_config)
        .should_receive("load_packit_yaml")
        .with_args(Path(res_pc_path))
        .and_return(None)
    )

    (
        flexmock(packit.config.package_config)
        .should_receive("parse_loaded_config")
        .and_return(None)
    )

    get_local_package_config(
        try_local_dir_last=local_last, package_config_path=config_file_name
    )


@pytest.mark.parametrize(
    "package_config",
    [
        PackageConfig(
            specfile_path="fedora/package.spec",
            downstream_package_name="package",
            upstream_package_name="package",
            synced_files=[
                SyncFilesItem(src=["fedora/package.spec"], dest="fedora/package.spec")
            ],
        ),
        PackageConfig(
            specfile_path="fedora/package.spec",
            downstream_package_name="package",
            synced_files=[
                SyncFilesItem(src=["fedora/package.spec"], dest="fedora/package.spec")
            ],
            jobs=[
                get_job_config_full(
                    specfile_path="fedora/package.spec",
                    downstream_package_name="package",
                    synced_files=[
                        SyncFilesItem(
                            src=["fedora/package.spec"], dest="fedora/package.spec"
                        )
                    ],
                )
            ],
        ),
        PackageConfig(
            specfile_path="fedora/package.spec",
            actions={
                ActionName.pre_sync: "some/pre-sync/command --option",
                ActionName.get_current_version: "get-me-version",
            },
            jobs=[],
            upstream_project_url="https://github.com/asd/qwe",
            upstream_package_name="qwe",
            dist_git_base_url="https://something.wicked",
            downstream_package_name="package",
            spec_source_id="Source1",
        ),
        PackageConfig(
            specfile_path="fedora/package.spec",
            actions={
                ActionName.pre_sync: "some/pre-sync/command --option",
                ActionName.get_current_version: "get-me-version",
            },
            jobs=[],
            upstream_project_url="https://github.com/asd/qwe",
            upstream_package_name="qwe",
            dist_git_base_url="https://something.wicked",
            downstream_package_name="package",
            spec_source_id="Source1",
            sources=[
                SourcesItem(
                    path="rsync-3.1.3.tar.gz",
                    url="https://git.centos.org/sources/rsync/c8s/82e7829",
                ),
            ],
        ),
    ],
)
def test_serialize_and_deserialize(package_config):
    schema = PackageConfigSchema()
    serialized = schema.dump(package_config)
    new_package_config = schema.load(serialized)
    assert package_config == new_package_config



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from pathlib import Path
from contextlib import contextmanager

import pytest

# from flexmock import flexmock

from packit.sync import check_subpath, SyncFilesItem
from packit.exceptions import PackitException


@contextmanager
def return_result(result):
    yield result


@pytest.mark.parametrize(
    "subpath,path,result",
    [
        (Path("./test/this"), Path("."), return_result(Path("./test/this").resolve())),
        (Path("test/this"), Path("."), return_result(Path("test/this").resolve())),
        (Path("../test/this"), Path("."), pytest.raises(PackitException)),
        (Path("test/../../this"), Path("."), pytest.raises(PackitException)),
    ],
)
def test_check_subpath(subpath, path, result):
    with result as r:
        assert check_subpath(subpath, path) == r


@pytest.mark.parametrize(
    "item,drop,result",
    [
        (SyncFilesItem(["a", "b"], "dest"), {"src": "b"}, SyncFilesItem(["a"], "dest")),
        (SyncFilesItem(["a"], "dest"), {"src": Path("a")}, None),
        (
            SyncFilesItem(["a", "b"], "dest"),
            {"src": "c"},
            SyncFilesItem(["a", "b"], "dest"),
        ),
        (
            SyncFilesItem(["src/a", "src/b"], "dest"),
            {"src": "a", "criteria": lambda x, y: x.name == y},
            SyncFilesItem(["src/b"], "dest"),
        ),
    ],
)
def test_drop_src(item, drop, result):
    """Check dropping a item from the src-list

    The 'src' argument can be a string or Path.
    When used in 'criteria' as 'y', there is no type conversion,
    so if the caller defines a custem 'criteria', it falls on them,
    to make sure the types of 'src' and 'y' match.
    """
    assert result == item.drop_src(**drop)


@pytest.mark.parametrize(
    "item,args,result",
    [
        (
            SyncFilesItem(["a", "b"], "dest"),
            {},
            return_result(
                SyncFilesItem(
                    [Path(Path.cwd() / "a"), Path(Path.cwd() / "b")],
                    Path(Path.cwd() / "dest"),
                )
            ),
        ),
        (
            SyncFilesItem(["a", "b"], "dest"),
            {"src_base": Path("src").resolve(), "dest_base": Path("dest")},
            return_result(
                SyncFilesItem(
                    [Path(Path.cwd() / "src" / "a"), Path(Path.cwd() / "src" / "b")],
                    Path(Path.cwd() / "dest" / "dest"),
                )
            ),
        ),
        (
            SyncFilesItem(["../a", "b"], "dest"),
            {"src_base": Path("src"), "dest_base": Path("dest").resolve()},
            pytest.raises(PackitException),
        ),
    ],
)
def test_resolve(item, args, result):
    with result as r:
        item.resolve(**args)
        assert r == item


@pytest.mark.parametrize(
    "item,args,result",
    [
        (
            SyncFilesItem(["a", "b"], "dest"),
            {},
            "rsync --archive --ignore-missing-args a b dest".split(),
        ),
        (
            SyncFilesItem(["a", "b"], "dest2"),
            {"fail_on_missing": True},
            "rsync --archive a b dest2".split(),
        ),
        (
            SyncFilesItem(["c", "d"], "dest", mkpath=True),
            {},
            "rsync --archive --mkpath --ignore-missing-args c d dest".split(),
        ),
        (
            SyncFilesItem(["c/*"], "dest", mkpath=True),
            {},
            "rsync --archive --mkpath --ignore-missing-args c/* dest".split(),
        ),
    ],
)
def test_command(item, args, result):
    assert result == item.command(**args)


def test_command_globs(tmp_path):
    src_dir = tmp_path / "src"
    src_dir.mkdir()
    (src_dir / "file1").touch()
    (src_dir / "file2").touch()
    item = SyncFilesItem(["src/*"], "dest")
    item.resolve(src_base=tmp_path, dest_base=tmp_path)
    command = item.command()
    # The return value of glob.glob() is unordered
    command = command[:-3] + sorted(command[-3:-1]) + command[-1:]
    assert [
        "rsync",
        "--archive",
        "--ignore-missing-args",
        f"{tmp_path}/src/file1",
        f"{tmp_path}/src/file2",
        f"{tmp_path}/dest",
    ] == command



# MIT License
#
# Copyright (c) 2020 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pathlib import Path
from typing import Dict, Union, List, Any

from marshmallow import ValidationError

from packit.config.package_config import PackageConfig, get_local_specfile_path
from packit.exceptions import PackitConfigException


class PackageConfigValidator:
    """validate content of .packit.yaml"""

    def __init__(self, config_file_path: Path, config_content: Dict):
        self.config_file_path = config_file_path
        self.content = config_content

    def validate(self) -> str:
        """Create output for PackageConfig validation."""
        schema_errors: Union[List[Any], Dict[Any, Any]] = None
        try:
            PackageConfig.get_from_dict(
                self.content,
                config_file_path=str(self.config_file_path),
                spec_file_path=str(
                    get_local_specfile_path(self.config_file_path.parent)
                ),
            )
        except ValidationError as e:
            schema_errors = e.messages
        except PackitConfigException as e:
            return str(e)

        if not schema_errors:
            return f"{self.config_file_path.name} is valid and ready to be used"

        output = f"{self.config_file_path.name} does not pass validation:\n"
        if isinstance(schema_errors, list):
            output += "\n".join(map(str, schema_errors))
            return output

        for field_name, errors in schema_errors.items():
            output += self.validate_get_field_output(errors, field_name)
        return output

    def validate_get_field_output(
        self,
        errors: Union[list, dict],
        field_name: str,
        field_category: str = "* field",
        level: int = 1,
    ) -> str:
        if isinstance(errors, list):
            field_output = f"{field_category} {field_name}: {errors[0]}\n"
            return field_output

        field_output = self.validate_get_field_item_output(errors, field_name, level)
        return field_output

    def validate_get_field_item_output(
        self, errors: dict, field_name: str, level: int
    ) -> str:
        index_output = f"{level * '*'} field {field_name} has an incorrect value:\n"
        level += 1
        for index, item_errors in errors.items():
            index_type_error = self.validate_get_field_output(
                item_errors,
                index,
                field_category=f"{level * '*'} value at index",
                level=level,
            )
            index_output += index_type_error
        return index_output



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import os

import click

from packit.local_project import LocalProject
from packit.utils.repo import git_remote_url_to_https_url

logger = logging.getLogger(__name__)


class LocalProjectParameter(click.ParamType):
    """
    Path or url.
    """

    name = "path_or_url"

    def __init__(self, branch_param_name: str = None) -> None:
        """
        :param branch_param_name: name of the cli function parameter (not the option name)
        """
        super().__init__()
        self.branch_param_name = branch_param_name

    def convert(self, value, param, ctx):
        try:
            branch_name = None
            if self.branch_param_name:
                if self.branch_param_name in ctx.params:
                    branch_name = ctx.params[self.branch_param_name]
                else:  # use the default
                    for param in ctx.command.params:
                        if param.name == self.branch_param_name:
                            branch_name = param.default

            if os.path.isdir(value):
                absolute_path = os.path.abspath(value)
                logger.debug(f"Input is a directory: {absolute_path}")
                local_project = LocalProject(
                    working_dir=absolute_path,
                    ref=branch_name,
                    remote=ctx.obj.upstream_git_remote,
                )
            elif git_remote_url_to_https_url(value):
                logger.debug(f"Input is a URL to a git repo: {value}")
                local_project = LocalProject(
                    git_url=value, ref=branch_name, remote=ctx.obj.upstream_git_remote
                )
            else:
                self.fail(
                    "Provided input path_or_url is not a directory nor an URL of a git repo."
                )

            if not (local_project.working_dir or local_project.git_url):
                self.fail(
                    "Parameter is not an existing directory nor correct git url.",
                    param,
                    ctx,
                )
            return local_project
        except Exception as ex:
            self.fail(ex, param, ctx)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from packit.utils.commands import run_command, run_command_remote, cwd
from packit.utils.extensions import assert_existence, nested_get
from packit.utils.logging import (
    StreamLogger,
    PackitFormatter,
    set_logging,
    commits_to_nice_str,
)
from packit.utils.repo import (
    is_git_repo,
    get_repo,
    get_namespace_and_repo_name,
    is_a_git_ref,
    git_remote_url_to_https_url,
)


__all__ = [
    run_command.__name__,
    run_command_remote.__name__,
    cwd.__name__,
    assert_existence.__name__,
    nested_get.__name__,
    StreamLogger.__name__,
    PackitFormatter.__name__,
    set_logging.__name__,
    commits_to_nice_str.__name__,
    is_git_repo.__name__,
    get_repo.__name__,
    get_namespace_and_repo_name.__name__,
    is_a_git_ref.__name__,
    git_remote_url_to_https_url.__name__,
]


def sanitize_branch_name(branch_name: str) -> str:
    """
    replace potentially problematic characters in provided string, a branch name

    e.g. copr says:
        Name must contain only letters, digits, underscores, dashes and dots.
    """
    # https://stackoverflow.com/questions/3411771/best-way-to-replace-multiple-characters-in-a-string
    offenders = "!@#$%^&*()+={[}]|\\'\":;<,>/?~`"
    for o in offenders:
        branch_name = branch_name.replace(o, "-")
    return branch_name


def sanitize_branch_name_for_rpm(branch_name: str) -> str:
    """
    rpm is picky about release: hates "/" - it's an error
    also prints a warning for "-"
    """
    offenders = "!@#$%^&*()+={[}]|\\'\":;<,>/?~`"
    for o in offenders:
        branch_name = branch_name.replace(o, "")
    return branch_name.replace("-", ".")



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
E2E tests which utilize cockpit projects
"""
import shutil
from pathlib import Path

import pytest
from flexmock import flexmock

from packit.api import PackitAPI
from packit.config import get_local_package_config
from packit.distgit import DistGit
from packit.fedpkg import FedPKG
from packit.local_project import LocalProject
from packit.utils import repo
from packit.utils.commands import cwd
from tests.spellbook import UP_COCKPIT_OSTREE, initiate_git_repo, get_test_config


@pytest.fixture()
def cockpit_ostree(tmp_path, upstream_without_config):
    u = tmp_path / "up"
    initiate_git_repo(u, tag="179", copy_from=UP_COCKPIT_OSTREE)

    flexmock(repo, get_namespace_and_repo_name=lambda url: ("asd", "qwe"))
    d = tmp_path / "dg"
    d.mkdir()
    initiate_git_repo(d, upstream_remote=upstream_without_config, push=True)

    shutil.copy2(
        UP_COCKPIT_OSTREE / "cockpit-ostree.spec.dg", d / "cockpit-ostree.spec"
    )

    return u, d


def test_update_on_cockpit_ostree(cockpit_ostree):
    upstream_path, dist_git_path = cockpit_ostree

    def mocked_new_sources(sources=None):
        if not Path(sources).is_file():
            raise RuntimeError("archive does not exist")

    flexmock(FedPKG, new_sources=mocked_new_sources)
    flexmock(PackitAPI, init_kerberos_ticket=lambda: None)

    flexmock(
        DistGit,
        push_to_fork=lambda *args, **kwargs: None,
        is_archive_in_lookaside_cache=lambda archive_path: False,
        upload_to_lookaside_cache=lambda path: None,
        download_upstream_archive=lambda: "the-archive",
    )
    flexmock(
        PackitAPI,
        push_and_create_pr=lambda pr_title, pr_description, dist_git_branch: None,
    )

    pc = get_local_package_config(str(upstream_path))
    up_lp = LocalProject(working_dir=upstream_path)
    c = get_test_config()
    api = PackitAPI(c, pc, up_lp)
    api._dg = DistGit(c, pc)
    api._dg._local_project = LocalProject(working_dir=dist_git_path)

    with cwd(upstream_path):
        api.sync_release(
            dist_git_branch="main",
            use_local_content=False,
            version="179",
            force_new_sources=False,
            create_pr=True,
        )


def test_srpm_on_cockpit_ostree(cockpit_ostree):
    upstream_path, dist_git_path = cockpit_ostree

    pc = get_local_package_config(str(upstream_path))
    up_lp = LocalProject(working_dir=upstream_path)
    c = get_test_config()
    api = PackitAPI(c, pc, up_lp)

    with cwd(upstream_path):
        api.create_srpm()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os
import subprocess
from pathlib import Path

import pytest

from packit.exceptions import PackitException
from packit.patches import PatchGenerator
from packit.specfile import Specfile
from packit.utils.commands import cwd
from tests.integration.conftest import mock_spec_download_remote_s
from tests.spellbook import (
    TARBALL_NAME,
    git_add_and_commit,
    build_srpm,
    create_merge_commit_in_source_git,
    create_git_am_style_history,
    create_patch_mixed_history,
    create_history_with_empty_commit,
    run_prep_for_srpm,
)


DISTRO_DIR = ".distro"


def test_basic_local_update_without_patching(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git"""

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)

    api_instance_source_git.sync_release(
        dist_git_branch="main",
        version="0.1.0",
        upstream_ref="0.1.0",
    )

    assert (distgit / TARBALL_NAME).is_file()
    spec = Specfile(distgit / "beer.spec")
    assert spec.get_version() == "0.1.0"


@pytest.mark.parametrize("ref", [None, "0.1.0", "0.1*", "0.*"])
def test_basic_local_update_empty_patch(
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
    ref,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git"""

    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)
    api_instance_source_git.sync_release(
        dist_git_branch="main",
        version="0.1.0",
        upstream_ref=ref,
    )

    assert (distgit / TARBALL_NAME).is_file()
    spec = Specfile(distgit / "beer.spec")
    assert spec.get_version() == "0.1.0"

    spec_package_section = ""
    for section in spec.spec_content.sections:
        if "%package" in section[0]:
            spec_package_section += "\n".join(section[1])
    assert not spec.patches["applied"]
    assert not spec.patches["not_applied"]


def test_basic_local_update_patch_content(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git"""

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)

    create_merge_commit_in_source_git(sourcegit)

    source_file = sourcegit / "big-source-file.txt"
    source_file.write_text("new changes")
    git_add_and_commit(directory=sourcegit, message="source change")

    source_file = sourcegit / "ignored_file.txt"
    source_file.write_text(" And I am sad.")
    git_add_and_commit(directory=sourcegit, message="make a file sad")

    api_instance_source_git.sync_release(
        dist_git_branch="main",
        version="0.1.0",
        upstream_ref="0.1.0",
    )

    git_diff = subprocess.check_output(
        ["git", "diff", "HEAD~", "HEAD"], cwd=distgit
    ).decode()

    assert (
        """
-Version:        0.0.0
+Version:        0.1.0"""
        in git_diff
    )
    # Make sure the patches are placed after Source0, but outside %if %endif
    patches = (
        """\
Source0:        %{upstream_name}-%{version}.tar.gz
 %endif
 BuildArch:      noarch
 """  # because linters would remove the trailing whitespace
        + """
+# switching to amarillo hops
+# Author: Packit Test Suite <test@example.com>
+Patch0001: 0001-switching-to-amarillo-hops.patch
+
+# actually, let's do citra
+# Author: Packit Test Suite <test@example.com>
+Patch0002: 0002-actually-let-s-do-citra.patch
+
+# source change
+# Author: Packit Test Suite <test@example.com>
+Patch0003: 0003-source-change.patch
+
"""
    )
    assert patches in git_diff

    assert "Patch0004:" not in git_diff

    assert (
        """ - 0.1.0-1
+- commit with data (Packit Test Suite)
+- empty commit #2 (Packit Test Suite)
+- empty commit #1 (Packit Test Suite)
+- empty commit #0 (Packit Test Suite)
+
 * Sun Feb 24 2019 Tomas Tomecek <ttomecek@redhat.com> - 0.0.0-1
 - No brewing, yet."""
        in git_diff
    )

    # direct diff in the synced file
    assert (
        """diff --git a/.packit.yaml b/.packit.yaml
new file mode 100644"""
        in git_diff
    )

    assert (
        """
--- /dev/null
+++ b/.packit.yaml"""
        in git_diff
    )

    # diff of the synced file should not be in the patch
    assert (
        """
+diff --git a/.packit.yaml b/.packit.yaml
+new file mode 100644"""
        not in git_diff
    )

    patch_1_3 = """
+Subject: [PATCH 1/3] switching to amarillo hops
+
+---
+ hops | 2 +-
+ 1 file changed, 1 insertion(+), 1 deletion(-)
+
+diff --git a/hops b/hops"""
    assert patch_1_3 in git_diff
    assert (
        """\
+--- a/hops
++++ b/hops
+@@ -1 +1 @@
+-Cascade
++Amarillo
+--"""
        in git_diff
    )

    assert (
        """\
+Subject: [PATCH 2/3] actually, let's do citra
+
+---
+ hops | 2 +-
+ 1 file changed, 1 insertion(+), 1 deletion(-)
+
+diff --git a/hops b/hops"""
        in git_diff
    )
    assert (
        (
            """\
+--- a/hops
++++ b/hops
+@@ -1 +1 @@
+-Amarillo
++Citra
+--"""
        )
        in git_diff
    )

    assert (
        """
+--- a/big-source-file.txt
++++ b/big-source-file.txt
+@@ -1,2 +1 @@
+-This is a testing file
+-containing some text.
++new changes"""
        in git_diff
    )

    # diff of the source files (not synced) should not be directly in the git diff
    assert (
        """
+Subject: [PATCH 3/3] source change
+
+---
+ big-source-file.txt | 3 +--
+ 1 file changed, 1 insertion(+), 2 deletions(-)
+
+diff --git a/big-source-file.txt b/big-source-file.txt"""
        in git_diff
    )

    # ignored file should not be in the diff
    assert "--- a/ignored_file.txt\n" not in git_diff


def test_basic_local_update_patch_content_with_metadata(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git"""

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)

    create_merge_commit_in_source_git(sourcegit)

    source_file = sourcegit / "big-source-file.txt"
    source_file.write_text("new changes")
    git_add_and_commit(
        directory=sourcegit,
        message="source change\n"
        "patch_name: testing.patch\n"
        "description: Few words for info.",
    )

    source_file = sourcegit / "ignored_file.txt"
    source_file.write_text(" And I am sad.")
    git_add_and_commit(directory=sourcegit, message="make a file sad")

    api_instance_source_git.sync_release(
        dist_git_branch="main",
        version="0.1.0",
        upstream_ref="0.1.0",
    )

    git_diff = subprocess.check_output(
        ["git", "diff", "HEAD~", "HEAD"], cwd=distgit
    ).decode()

    patches = """
+# switching to amarillo hops
+# Author: Packit Test Suite <test@example.com>
+Patch0001: 0001-switching-to-amarillo-hops.patch
+
+# actually, let's do citra
+# Author: Packit Test Suite <test@example.com>
+Patch0002: 0002-actually-let-s-do-citra.patch
+
+# source change
+# Author: Packit Test Suite <test@example.com>
+# Few words for info.
+Patch0003: testing.patch
+
"""
    assert patches in git_diff


def test_basic_local_update_patch_content_with_metadata_and_patch_ignored(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git"""

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)

    create_merge_commit_in_source_git(sourcegit)

    source_file = sourcegit / "big-source-file.txt"
    source_file.write_text("new changes")
    git_add_and_commit(
        directory=sourcegit,
        message="source change\nignore: true",
    )

    source_file = sourcegit / "ignored_file.txt"
    source_file.write_text(" And I am sad.")
    git_add_and_commit(directory=sourcegit, message="make a file sad")

    api_instance_source_git.sync_release(
        dist_git_branch="main",
        version="0.1.0",
        upstream_ref="0.1.0",
    )

    git_diff = subprocess.check_output(
        ["git", "diff", "HEAD~", "HEAD"], cwd=distgit
    ).decode()

    patches = """
+# switching to amarillo hops
+# Author: Packit Test Suite <test@example.com>
+Patch0001: 0001-switching-to-amarillo-hops.patch
+
+# actually, let's do citra
+# Author: Packit Test Suite <test@example.com>
+Patch0002: 0002-actually-let-s-do-citra.patch
+
"""
    assert patches in git_diff


def test_basic_local_update_patch_content_with_downstream_patch(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git"""

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)

    create_merge_commit_in_source_git(sourcegit)

    source_file = sourcegit / "ignored_file.txt"
    source_file.write_text(" And I am sad.")
    git_add_and_commit(directory=sourcegit, message="make a file sad")

    api_instance_source_git.sync_release(
        dist_git_branch="main",
        version="0.1.0",
        upstream_ref="0.1.0",
    )

    git_diff = subprocess.check_output(
        ["git", "diff", "HEAD~", "HEAD"], cwd=distgit
    ).decode()

    patches = """
+# switching to amarillo hops
+# Author: Packit Test Suite <test@example.com>
+Patch0001: 0001-switching-to-amarillo-hops.patch
+
+# actually, let's do citra
+# Author: Packit Test Suite <test@example.com>
+Patch0002: 0002-actually-let-s-do-citra.patch
+
"""
    assert patches in git_diff


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm(mock_remote_functionality_sourcegit, api_instance_source_git, ref):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")
    create_merge_commit_in_source_git(sg_path)
    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)
    srpm_path = list(sg_path.glob("beer-0.1.0-2.*.src.rpm"))[0]
    assert srpm_path.is_file()
    build_srpm(srpm_path)
    branches = subprocess.check_output(
        ["git", "for-each-ref", "--format=%(refname:short)", "refs/heads/"], cwd=sg_path
    ).split(b"\n")
    for b in branches:
        if b and b.startswith(b"packit-patches-"):
            raise AssertionError(
                "packit-patches- branch was found - the history shouldn't have been linearized"
            )
    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "0001-switching-to-amarillo-hops.patch",
        "0002-actually-let-s-do-citra.patch",
    }


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm_merge_storm(
    mock_remote_functionality_sourcegit, api_instance_source_git, ref
):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")
    create_merge_commit_in_source_git(sg_path, go_nuts=True)

    # linearization creates a new branch, make some arbitrary moves to verify
    # we end up in the former branch after the build
    subprocess.check_call(["git", "checkout", "-B", "test-branch"], cwd=sg_path)
    subprocess.check_call(["git", "checkout", "main"], cwd=sg_path)

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)
    srpm_path = list(sg_path.glob("beer-0.1.0-2.*.src.rpm"))[0]
    assert srpm_path.is_file()
    build_srpm(srpm_path)
    branches = subprocess.check_output(
        ["git", "for-each-ref", "--format=%(refname:short)", "refs/heads/"], cwd=sg_path
    ).split(b"\n")
    for b in branches:
        if b and b.startswith(b"packit-patches-"):
            break
    else:
        raise AssertionError(
            "packit-patches- branch was not found - this should trigger the linearization"
        )
    # make sure we are on the main branch
    assert (
        "main"
        == subprocess.check_output(["git", "branch", "--show-current"], cwd=sg_path)
        .decode()
        .strip()
    )
    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "0001-MERGE-COMMIT.patch",
        "0002-ugly-merge-commit.patch",
    }


def test_srpm_merge_storm_dirty(api_instance_source_git):
    """verify the linearization is halted when a source-git repo si dirty"""
    ref = "0.1.0"
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, ref)
    create_merge_commit_in_source_git(sg_path, go_nuts=True)
    (sg_path / "malt").write_text("Mordor\n")
    with pytest.raises(PackitException) as ex:
        with cwd("/"):  # let's mimic p-s by having different cwd than the project
            api_instance_source_git.create_srpm(upstream_ref=ref)
    assert "The source-git repo is dirty" in str(ex.value)


def test_linearization(api_instance_source_git):
    ref = "0.1.0"
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, ref)
    create_merge_commit_in_source_git(sg_path, go_nuts=True)
    with cwd("/"):  # let's mimic p-s by having different cwd than the project
        pg = PatchGenerator(api_instance_source_git.upstream_local_project)
        pg.create_patches(ref, sg_path / DISTRO_DIR)
    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "0001-sourcegit-content.patch",
        "0002-MERGE-COMMIT.patch",
        "0003-ugly-merge-commit.patch",
    }


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm_git_am(mock_remote_functionality_sourcegit, api_instance_source_git, ref):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")

    api_instance_source_git.up.specfile.spec_content.section("%package")[10:10] = (
        "Patch1: citra.patch",
        "Patch2: malt.patch",
        "Patch8: 0001-m04r-malt.patch",
    )
    autosetup_line = api_instance_source_git.up.specfile.spec_content.section("%prep")[
        0
    ]
    autosetup_line = autosetup_line.replace("-S patch", "-S git_am")
    api_instance_source_git.up.specfile.spec_content.section("%prep")[
        0
    ] = autosetup_line
    api_instance_source_git.up.specfile.save()

    create_git_am_style_history(sg_path)

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)

    srpm_path = list(sg_path.glob("beer-0.1.0-2.*.src.rpm"))[0]
    assert srpm_path.is_file()
    build_srpm(srpm_path)

    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "citra.patch",
        "0001-m04r-malt.patch",
        "malt.patch",
    }
    run_prep_for_srpm(srpm_path)
    prep_root = sg_path.joinpath("beerware-0.1.0")
    assert prep_root.joinpath("malt").read_text() == "Munich\n"
    assert prep_root.joinpath("hops").read_text() == "Saaz\n"


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm_git_no_prefix_patches(
    mock_remote_functionality_sourcegit, api_instance_source_git, ref
):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")

    api_instance_source_git.up.specfile.spec_content.section("%package")[10:10] = (
        "Patch1: amarillo.patch",
        "Patch2: citra.patch",
        "Patch8: malt.patch",
    )
    api_instance_source_git.up.specfile.spec_content.section("%prep")[0:2] = [
        "%setup -n %{upstream_name}-%{version}",
        "%patch1 -p1",
        "%patch2 -p0",
        "%patch8 -p1",
    ]
    api_instance_source_git.up.specfile.save()

    create_patch_mixed_history(sg_path)

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)

    srpm_path = list(sg_path.glob("beer-0.1.0-2.*.src.rpm"))[0]
    assert srpm_path.is_file()
    build_srpm(srpm_path)

    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "amarillo.patch",
        "citra.patch",
        "malt.patch",
    }


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm_empty_patch(
    mock_remote_functionality_sourcegit, api_instance_source_git, ref
):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")

    api_instance_source_git.up.specfile.spec_content.section("%package")[10:10] = (
        "Patch1: amarillo.patch",
        "Patch2: citra.patch",
        "Patch5: saaz.patch",
        "Patch8: malt.patch",
    )
    api_instance_source_git.up.specfile.save()

    create_history_with_empty_commit(sg_path)

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)

    srpm_path = list(sg_path.glob("beer-0.1.0-2.*.src.rpm"))[0]
    assert srpm_path.is_file()
    build_srpm(srpm_path)

    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "amarillo.patch",
        "citra.patch",
        "saaz.patch",
        "malt.patch",
    }
    assert sg_path.joinpath(DISTRO_DIR, "saaz.patch").read_text() == ""


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm_patch_non_conseq_indices(
    mock_remote_functionality_sourcegit, api_instance_source_git, ref
):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")

    api_instance_source_git.up.specfile.spec_content.section("%package")[10:10] = (
        "Patch0: amarillo.patch",
        "Patch3: citra.patch",
        "Patch4: saaz.patch",
        "Patch5: malt.patch",
    )
    api_instance_source_git.up.specfile.save()

    create_history_with_empty_commit(sg_path)

    malt = sg_path.joinpath("malt")
    malt.write_text("Wheat\n")
    git_add_and_commit(directory=sg_path, message="Weißbier! Summer is coming!")

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)

    # make sure the patch is inserted AFTER existing patches
    patches = api_instance_source_git.up.specfile.tags.filter(name="Patch*", valid=None)
    last_patch = list(patches)[-1]
    assert last_patch.name == "Patch6"
    assert (
        os.path.basename(
            api_instance_source_git.up.specfile.get_applied_patches()[-1].path
        )
        == "0004-Wei-bier-Summer-is-coming.patch"
    )

    srpm_path = list(sg_path.glob("beer-0.1.0-2.*.src.rpm"))[0]
    assert srpm_path.is_file()
    build_srpm(srpm_path)

    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "amarillo.patch",
        "citra.patch",
        "saaz.patch",
        "malt.patch",
        "0004-Wei-bier-Summer-is-coming.patch",
    }
    assert sg_path.joinpath(DISTRO_DIR, "saaz.patch").read_text() == ""



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
from git import PushInfo

from flexmock import flexmock
from packit.actions import ActionName
from packit.base_git import PackitRepositoryBase
from packit.exceptions import PackitException
from packit.config import Config, PackageConfig, RunCommandType
from packit.local_project import LocalProject
from tests.spellbook import can_a_module_be_imported


@pytest.mark.parametrize(
    "echo_cmd, expected_output",
    [
        ("echo 'hello world'", ["hello world\n"]),
        # should return output of only the last one
        (["echo 'ignore me'", "echo 'hello world'"], ["ignore me\n", "hello world\n"]),
    ],
)
def test_get_output_from_action_defined(echo_cmd, expected_output):
    packit_repository_base = PackitRepositoryBase(
        config=flexmock(Config()),
        package_config=flexmock(PackageConfig(actions={ActionName.pre_sync: echo_cmd})),
    )

    packit_repository_base.local_project = flexmock(working_dir=".")

    result = packit_repository_base.get_output_from_action(ActionName.pre_sync)
    assert result == expected_output


@pytest.mark.skipif(
    not can_a_module_be_imported("sandcastle"), reason="sandcastle is not installed"
)
def test_get_output_from_action_defined_in_sandcastle():
    from sandcastle.api import Sandcastle

    echo_cmd = "hello world"
    flexmock(Sandcastle).should_receive("get_api_client")
    flexmock(Sandcastle).should_receive("is_pod_already_deployed").and_return(True)
    c = Config()
    c.command_handler = RunCommandType.sandcastle
    packit_repository_base = PackitRepositoryBase(
        config=c, package_config=PackageConfig(actions={ActionName.pre_sync: echo_cmd})
    )
    packit_repository_base.local_project = LocalProject(working_dir="/tmp")

    flexmock(Sandcastle).should_receive("run")
    flexmock(Sandcastle).should_receive("exec").and_return(echo_cmd)
    flexmock(Sandcastle).should_receive("delete_pod").and_return(None)
    result = packit_repository_base.get_output_from_action(ActionName.pre_sync)
    assert result[-1] == echo_cmd


@pytest.mark.skip(
    reason="Skipping since we don't have an OpenShift cluster by default."
)
def test_run_in_sandbox():
    packit_repository_base = PackitRepositoryBase(
        config=Config(),
        package_config=PackageConfig(actions={ActionName.pre_sync: "ls -lha"}),
    )
    packit_repository_base.config.actions_handler = "sandcastle"

    result = packit_repository_base.get_output_from_action(ActionName.pre_sync)
    assert "total 4.0K" in result
    assert "drwxr-xr-x. 1 root root" in result


def test_base_push_bad(distgit_and_remote):
    distgit, _ = distgit_and_remote

    b = PackitRepositoryBase(config=Config(), package_config=PackageConfig())
    b.local_project = LocalProject(
        working_dir=distgit, git_url="https://github.com/packit/lol"
    )
    flexmock(
        LocalProject,
        push=lambda *args, **kwargs: [
            PushInfo(PushInfo.REMOTE_REJECTED, None, None, None, None)
        ],
    )
    with pytest.raises(PackitException) as e:
        b.push("master")
    assert "unable to push" in str(e.value)


def test_base_push_good(distgit_and_remote):
    distgit, _ = distgit_and_remote

    b = PackitRepositoryBase(config=Config(), package_config=PackageConfig())
    b.local_project = LocalProject(
        working_dir=distgit, git_url="https://github.com/packit/lol"
    )
    flexmock(
        LocalProject,
        push=lambda *args, **kwargs: [
            PushInfo(PushInfo.FAST_FORWARD, None, None, None, None)
        ],
    )
    b.push("master")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Common package config attributes so they can be imported both in PackageConfig and JobConfig
"""
import os
from typing import Dict, List, Optional, Union
from pathlib import Path

from packit.actions import ActionName
from packit.config.notifications import (
    NotificationsConfig,
    PullRequestNotificationsConfig,
)
from packit.config.sources import SourcesItem
from packit.constants import PROD_DISTGIT_URL
from packit.sync import SyncFilesItem, iter_srcs


class CommonPackageConfig:
    """
    We want JobConfig to hold all the attributes from PackageConfig so we don't need to
    pass both PackageConfig and JobConfig to handlers in p-s. We also want people
    to be able to override global PackageConfig attributes per job.

                        CommonPackageConfig
                              /      \
                   PackageConfig   JobConfig
                          //
              contains [JobConfig]
    """

    def __init__(
        self,
        config_file_path: Optional[str] = None,
        specfile_path: Optional[str] = None,
        synced_files: Optional[List[SyncFilesItem]] = None,
        dist_git_namespace: str = None,
        upstream_project_url: str = None,  # can be URL or path
        upstream_package_name: str = None,
        downstream_project_url: str = None,
        downstream_package_name: str = None,
        dist_git_base_url: str = None,
        create_tarball_command: List[str] = None,
        current_version_command: List[str] = None,
        actions: Dict[ActionName, Union[str, List[str]]] = None,
        upstream_ref: Optional[str] = None,
        allowed_gpg_keys: Optional[List[str]] = None,
        create_pr: bool = True,
        sync_changelog: bool = False,
        spec_source_id: str = "Source0",
        upstream_tag_template: str = "{version}",
        archive_root_dir_template: str = "{upstream_pkg_name}-{version}",
        patch_generation_ignore_paths: List[str] = None,
        notifications: Optional[NotificationsConfig] = None,
        copy_upstream_release_description: bool = False,
        sources: Optional[List[SourcesItem]] = None,
    ):
        self.config_file_path: Optional[str] = config_file_path
        self.specfile_path: Optional[str] = specfile_path
        self.synced_files: List[SyncFilesItem] = synced_files or []
        self.patch_generation_ignore_paths = patch_generation_ignore_paths or []
        self.dist_git_namespace: str = dist_git_namespace or "rpms"
        self.upstream_project_url: Optional[str] = upstream_project_url
        self.upstream_package_name: Optional[str] = upstream_package_name
        # this is generated by us
        self.downstream_package_name: Optional[str] = downstream_package_name
        self.dist_git_base_url: str = dist_git_base_url or os.getenv(
            "DISTGIT_URL", PROD_DISTGIT_URL
        )
        self._downstream_project_url: str = downstream_project_url
        # path to a local git clone of the dist-git repo; None means to clone in a tmpdir
        self.dist_git_clone_path: Optional[str] = None
        self.actions = actions or {}
        self.upstream_ref: Optional[str] = upstream_ref
        self.allowed_gpg_keys = allowed_gpg_keys
        self.create_pr: bool = create_pr
        self.sync_changelog: bool = sync_changelog
        self.spec_source_id: str = spec_source_id
        self.notifications = notifications or NotificationsConfig(
            pull_request=PullRequestNotificationsConfig()
        )

        # command to generate a tarball from the upstream repo
        # uncommitted changes will not be present in the archive
        self.create_tarball_command: List[str] = create_tarball_command
        # command to get current version of the project
        self.current_version_command: List[str] = current_version_command
        # template to create an upstream tag name (upstream may use different tagging scheme)
        self.upstream_tag_template = upstream_tag_template
        self.archive_root_dir_template = archive_root_dir_template
        self.copy_upstream_release_description = copy_upstream_release_description
        self.sources = sources or []

    def __repr__(self):
        return (
            "CommonPackageConfig("
            f"specfile_path='{self.specfile_path}', "
            f"synced_files='{self.synced_files}', "
            f"dist_git_namespace='{self.dist_git_namespace}', "
            f"upstream_project_url='{self.upstream_project_url}', "
            f"upstream_package_name='{self.upstream_package_name}', "
            f"downstream_project_url='{self.downstream_project_url}', "
            f"downstream_package_name='{self.downstream_package_name}', "
            f"dist_git_base_url='{self.dist_git_base_url}', "
            f"create_tarball_command='{self.create_tarball_command}', "
            f"current_version_command='{self.current_version_command}', "
            f"actions='{self.actions}', "
            f"upstream_ref='{self.upstream_ref}', "
            f"allowed_gpg_keys='{self.allowed_gpg_keys}', "
            f"create_pr='{self.create_pr}', "
            f"synced_files='{self.synced_files}', "
            f"spec_source_id='{self.spec_source_id}', "
            f"upstream_tag_template='{self.upstream_tag_template}', "
            f"patch_generation_ignore_paths='{self.patch_generation_ignore_paths}',"
            f"copy_upstream_release_description='{self.copy_upstream_release_description}',"
            f"sources='{self.sources}')"
        )

    @property
    def downstream_project_url(self) -> str:
        if not self._downstream_project_url:
            self._downstream_project_url = self.dist_git_package_url
        return self._downstream_project_url

    @property
    def dist_git_package_url(self):
        return (
            f"{self.dist_git_base_url}{self.dist_git_namespace}/"
            f"{self.downstream_package_name}.git"
        )

    def get_specfile_sync_files_item(self):
        """
        Get SyncFilesItem object for the specfile.
        :return: SyncFilesItem
        """
        return SyncFilesItem(
            src=[self.specfile_path],
            dest=f"{self.downstream_package_name}.spec"
            if self.downstream_package_name
            else self.specfile_path,
        )

    def get_all_files_to_sync(self):
        """
        Adds the default files (config file, spec file) to synced files
        when doing propose-downstream.
        :return: Files to be synced
        """
        files = self.synced_files

        if Path(self.specfile_path) not in iter_srcs(files):
            files.append(self.get_specfile_sync_files_item())

        if Path(self.config_file_path) and self.config_file_path not in iter_srcs(
            files
        ):
            # this relative because of glob: "Non-relative patterns are unsupported"
            files.append(
                SyncFilesItem(src=[self.config_file_path], dest=self.config_file_path)
            )

        return files



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys

from flexmock import flexmock

from packit.cli.utils import cover_packit_exception
from packit.exceptions import PackitException


def test_cover_packit_exception_decorator():
    flexmock(sys).should_receive("exit")

    @cover_packit_exception
    def covered_func():
        raise PackitException("Test exception")

    covered_func()


def test_cover_packit_exception_decorator_attribute():
    flexmock(sys).should_receive("exit")

    @cover_packit_exception
    def covered_func(attribute):
        assert attribute
        raise PackitException("Other test exception")

    covered_func(attribute=flexmock())


def test_cover_packit_exception_decorator_exit_code_default():
    flexmock(sys).should_receive("exit").with_args(2)

    @cover_packit_exception()
    def covered_func():
        raise PackitException("Test exception")

    covered_func()


def test_cover_packit_exception_decorator_exit_code_keyboard_interrupt():
    flexmock(sys).should_receive("exit").with_args(1)

    @cover_packit_exception()
    def covered_func():
        raise KeyboardInterrupt("Test keyboard interrupt")

    covered_func()


def test_cover_packit_exception_decorator_exit_code_other_exception():
    flexmock(sys).should_receive("exit").with_args(4)

    class CustomException(Exception):
        pass

    @cover_packit_exception()
    def covered_func():
        raise CustomException("Test keyboard interrupt")

    covered_func()


def test_cover_packit_exception_decorator_exit_code_other_exception_override():
    flexmock(sys).should_receive("exit").with_args(5)

    class CustomException(Exception):
        pass

    @cover_packit_exception(exit_code=5)
    def covered_func():
        raise CustomException("Test keyboard interrupt")

    covered_func()


def test_cover_packit_exception_decorator_exit_code_override():
    flexmock(sys).should_receive("exit").with_args(5)

    @cover_packit_exception(exit_code=5)
    def covered_func():
        raise PackitException("Test exception")

    covered_func()


def test_cover_packit_exception_decorator_config_debug_false():
    flexmock(sys).should_receive("exit")

    @cover_packit_exception
    def covered_func(config=None):
        raise PackitException("Test exception")

    covered_func(config=flexmock(debug=False))


def test_cover_packit_exception_decorator_config_debug_true():
    flexmock(sys).should_receive("exit")

    @cover_packit_exception
    def covered_func(config=None):
        raise PackitException("Test exception")

    covered_func(config=flexmock(debug=True))


def test_cover_packit_exception_decorator_other_exception():
    flexmock(sys).should_receive("exit").with_args(4)

    class CustomException(Exception):
        pass

    @cover_packit_exception
    def covered_func():
        raise CustomException("Other test exception")

    covered_func()


def test_cover_packit_exception_decorator_other_exception_config_debug():
    flexmock(sys).should_receive("exit")

    class CustomException(Exception):
        pass

    @cover_packit_exception
    def covered_func(config=None):
        raise CustomException("Other test exception")

    covered_func(config=flexmock(debug=True))



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from datetime import datetime, timedelta
from typing import List, Tuple, Dict, Set

from koji import ClientSession, BUILD_STATES

from ogr.abstract import Release
from packit.config import Config
from packit.config.common_package_config import CommonPackageConfig
from packit.copr_helper import CoprHelper
from packit.distgit import DistGit
from packit.exceptions import PackitException
from packit.upstream import Upstream

logger = logging.getLogger(__name__)


class Status:
    """
    This class provides methods to obtain status of the package
    """

    def __init__(
        self,
        config: Config,
        package_config: CommonPackageConfig,
        upstream: Upstream,
        distgit: DistGit,
    ):
        self.config = config
        self.package_config = package_config

        self.up = upstream
        self.dg = distgit

    def get_downstream_prs(self, number_of_prs: int = 5) -> List[Tuple[int, str, str]]:
        """
        Get specific number of latest downstream PRs
        :param number_of_prs: int
        :return: List of downstream PRs
        """
        table: List[Tuple[int, str, str]] = []
        pr_list = self.dg.local_project.git_project.get_pr_list()
        logger.debug("Downstream PRs fetched.")
        if len(pr_list) > 0:
            # take last `number_of_prs` PRs
            pr_list = (
                pr_list[:number_of_prs] if len(pr_list) > number_of_prs else pr_list
            )
            table = [(pr.id, pr.title, pr.url) for pr in pr_list]
        return table

    def get_dg_versions(self) -> Dict:
        """
        Get versions from all branches in Dist-git
        :return: Dict {"branch": "version"}
        """
        dg_versions = {}
        branches = self.dg.local_project.git_project.get_branches()
        logger.debug("Dist-git branches fetched.")
        for branch in branches:
            try:
                self.dg.create_branch(
                    branch, base=f"remotes/origin/{branch}", setup_tracking=False
                )
                self.dg.checkout_branch(branch)
                self.dg.specfile.update_spec()
            except Exception as ex:
                logger.debug(f"Branch {branch!r} is not present: {ex!r}.")
                continue
            try:
                dg_versions[branch] = self.dg.specfile.get_version()
            except PackitException:
                logger.debug(f"Can't figure out the version of branch: {branch}.")
        self.dg.checkout_branch()

        return dg_versions

    def get_up_releases(self, number_of_releases: int = 5) -> List:
        """
        Get specific number of latest upstream releases
        :param number_of_releases: int
        :return: List
        """
        if self.up.local_project.git_project is None:
            logger.info("We couldn't track any upstream releases.")
            return []

        latest_releases: List[Release] = []
        try:
            latest_releases = self.up.local_project.git_project.get_releases()
            logger.debug("Upstream releases fetched.")
        except PackitException as e:
            logger.debug(f"Failed to fetch upstream releases: {e}")

        return latest_releases[:number_of_releases]

    def get_koji_builds(
        self,
    ) -> Dict:
        """
        Get latest koji builds as a dict of branch: latest build in that branch.
        """
        session = ClientSession(baseurl="https://koji.fedoraproject.org/kojihub")
        package_id = session.getPackageID(
            self.dg.package_config.downstream_package_name
        )
        # This method returns only latest builds,
        # so we don't need to get whole build history from Koji,
        # get just recent year to speed things up.
        since = datetime.now() - timedelta(days=365)
        builds_l = session.listBuilds(
            packageID=package_id,
            state=BUILD_STATES["COMPLETE"],
            completeAfter=since.timestamp(),
        )
        logger.debug(f"Recent Koji builds fetched: {[b['nvr'] for b in builds_l]}")
        # Select latest build for each branch.
        # [{'nvr': 'python-ogr-0.5.0-1.fc29'}, {'nvr':'python-ogr-0.6.0-1.fc29'}]
        # -> {'fc29': 'python-ogr-0.6.0-1.fc29'}
        return {b["nvr"].rsplit(".", 1)[1]: b["nvr"] for b in reversed(builds_l)}

    def get_updates(self, number_of_updates: int = 3) -> List:
        """
        Get specific number of latest updates in bodhi
        :param number_of_updates: int
        :return: None
        """
        # https://github.com/fedora-infra/bodhi/issues/3058
        from bodhi.client.bindings import BodhiClient

        b = BodhiClient()
        results = b.query(packages=self.dg.package_config.downstream_package_name)[
            "updates"
        ]
        logger.debug("Bodhi updates fetched.")

        stable_branches: Set[str] = set()
        all_updates = [
            [
                result["title"],
                result["karma"],
                result["status"],
                result["release"]["branch"],
            ]
            for result in results
        ]
        updates = []
        for [update, karma, status, branch] in all_updates:
            # Don't return more than one stable update per branch
            if branch not in stable_branches or status != "stable":
                updates.append([update, karma, status])
                if status == "stable":
                    stable_branches.add(branch)
            if len(updates) == number_of_updates:
                break
        return updates

    def get_copr_builds(self, number_of_builds: int = 5) -> List:
        return CoprHelper(upstream_local_project=self.up.local_project).get_copr_builds(
            number_of_builds=number_of_builds
        )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
"""
packit started as source-git and we're making a source-git module after such a long time, weird
"""
import logging
import os
import shutil
import tarfile
import tempfile
from pathlib import Path
from typing import Optional, Tuple, List, Dict, Union
import configparser

import yaml
from git import GitCommandError

from rebasehelper.helpers.lookaside_cache_helper import LookasideCacheHelper
from rebasehelper.exceptions import LookasideCacheError


from packit.config.common_package_config import CommonPackageConfig
from packit.config.config import Config
from packit.config.package_config import PackageConfig
from packit.constants import (
    RPM_MACROS_FOR_PREP,
    FEDORA_DOMAIN,
    CENTOS_DOMAIN,
    CENTOS_STREAM_GITLAB,
    CENTOS_STREAM_GITLAB_DOMAIN,
    CENTOS_STREAM_GITLAB_NAMESPACE,
)
from packit.distgit import DistGit
from packit.exceptions import PackitException
from packit.local_project import LocalProject
from packit.utils.commands import run_command
from packit.utils.repo import (
    clone_centos_8_package,
    clone_centos_9_package,
    get_default_branch,
)

from ogr.parsing import parse_git_repo

logger = logging.getLogger(__name__)


class CentOS8DistGit(DistGit):
    """
    CentOS dist-git layout implementation for 8: CentOS Linux 8 and CentOS Stream 8
    which lives in git.centos.org
    """

    # spec files are stored in this dir in dist-git
    spec_dir_name = "SPECS"

    # sources are stored in this dir in dist-git
    # this applies to CentOS Stream 8 and CentOS Linux 7 and 8
    source_dir_name = "SOURCES"

    @classmethod
    def clone(
        cls,
        config: Config,
        package_config: CommonPackageConfig,
        path: Path,
        branch: str = None,
    ) -> "CentOS8DistGit":
        clone_centos_8_package(
            package_config.downstream_package_name, path, branch=branch
        )
        lp = LocalProject(working_dir=path)
        return cls(config, package_config, local_project=lp)


class CentOS9DistGit(DistGit):
    """
    CentOS dist-git layout implementation for CentOS Stream 9
    which lives in gitlab.com/redhat/centos-stream/rpms
    """

    # spec files are stored in this dir in dist-git
    spec_dir_name = ""

    # sources are stored in this dir in dist-git
    source_dir_name = ""

    @classmethod
    def clone(
        cls,
        config: Config,
        package_config: CommonPackageConfig,
        path: Path,
        branch: str = None,
    ) -> "CentOS9DistGit":
        clone_centos_9_package(
            package_config.downstream_package_name, path, branch=branch
        )
        lp = LocalProject(working_dir=path)
        return cls(config, package_config, local_project=lp)


def get_distgit_kls_from_repo(
    repo_path: Path, config: Config
) -> Tuple[DistGit, Optional[str], Optional[str]]:
    """
    :return: DistGit instance, centos package name, fedora package name
    """
    path = Path(repo_path)
    pc = PackageConfig(downstream_package_name=path.name)
    lp = LocalProject(working_dir=path)

    lp_git_repo = parse_git_repo(lp.git_url)

    if FEDORA_DOMAIN in lp_git_repo.hostname:
        return DistGit(config, pc, local_project=lp), None, path.name
    elif CENTOS_DOMAIN in lp_git_repo.hostname:
        return CentOS8DistGit(config, pc, local_project=lp), path.name, None
    elif (
        CENTOS_STREAM_GITLAB_DOMAIN == lp_git_repo.hostname
        and lp_git_repo.namespace.find(CENTOS_STREAM_GITLAB_NAMESPACE) == 0
    ):
        return CentOS9DistGit(config, pc, local_project=lp), path.name, None
    raise PackitException(
        f"Dist-git URL {lp.git_url} not recognized, we expected one of: "
        f"{FEDORA_DOMAIN}, {CENTOS_DOMAIN} or {CENTOS_STREAM_GITLAB}"
    )


def get_tarball_comment(tarball_path: str) -> Optional[str]:
    """Return the comment header for the tarball

    If written by git-archive, this contains the Git commit ID.
    Return None if the file is invalid or does not contain a comment.

    shamelessly stolen:
    https://pagure.io/glibc-maintainer-scripts/blob/master/f/glibc-sync-upstream.py#_75
    """
    try:
        with tarfile.open(tarball_path) as tar:
            return tar.pax_headers["comment"]
    except Exception as ex:
        logger.debug(f"Could not get 'comment' header from the tarball: {ex}")
        return None


# https://stackoverflow.com/questions/13518819/avoid-references-in-pyyaml
# mypy hated the suggestion from the SA ^, hence an override like this
class SafeDumperWithoutAliases(yaml.SafeDumper):
    def ignore_aliases(self, data):
        # no aliases/anchors in the dumped yaml text
        return True


class SourceGitGenerator:
    """
    generate a source-git repo from provided upstream repo
    and a corresponding package in Fedora/CentOS ecosystem
    """

    def __init__(
        self,
        local_project: LocalProject,
        config: Config,
        upstream_url: Optional[str] = None,
        upstream_ref: Optional[str] = None,
        dist_git_path: Optional[Path] = None,
        dist_git_branch: Optional[str] = None,
        fedora_package: Optional[str] = None,
        centos_package: Optional[str] = None,
        tmpdir: Optional[Path] = None,
    ):
        """
        :param local_project: this source-git repo
        :param config: global configuration
        :param upstream_url: upstream repo URL we want to use as a base
        :param upstream_ref: upstream git-ref to use as a base
        :param dist_git_path: path to a local clone of a dist-git repo
        :param dist_git_branch: branch in dist-git to use
        :param fedora_package: pick up specfile and downstream sources from this fedora package
        :param centos_package: pick up specfile and downstream sources from this centos package
        :param tmpdir: path to a directory where temporary repos (upstream,
                       dist-git) will be cloned
        """
        self.local_project = local_project
        self.config = config
        self.tmpdir = tmpdir or Path(tempfile.mkdtemp(prefix="packit-sg-"))
        self._dist_git: Optional[DistGit] = None
        self._primary_archive: Optional[Path] = None
        self._upstream_ref: Optional[str] = upstream_ref
        self.dist_git_branch = dist_git_branch

        logger.info(
            f"The source-git repo is going to be created in {local_project.working_dir}."
        )

        if dist_git_path:
            (
                self._dist_git,
                self.centos_package,
                self.fedora_package,
            ) = get_distgit_kls_from_repo(dist_git_path, config)
            self.dist_git_path = dist_git_path
            self.package_config = self.dist_git.package_config
        else:
            self.centos_package = centos_package
            self.fedora_package = fedora_package
            if centos_package:
                self.package_config = PackageConfig(
                    downstream_package_name=centos_package
                )
            elif fedora_package:
                self.fedora_package = (
                    self.fedora_package or local_project.working_dir.name
                )
                self.package_config = PackageConfig(
                    downstream_package_name=fedora_package
                )
            else:
                raise PackitException(
                    "Please tell us the name of the package in the downstream."
                )
            self.dist_git_path = self.tmpdir.joinpath(
                self.package_config.downstream_package_name
            )

        if upstream_url:
            if Path(upstream_url).is_dir():
                self.upstream_repo_path: Path = Path(upstream_url)
                self.upstream_lp: LocalProject = LocalProject(
                    working_dir=self.upstream_repo_path
                )
            else:
                self.upstream_repo_path = self.tmpdir.joinpath(
                    f"{self.package_config.downstream_package_name}-upstream"
                )
                self.upstream_lp = LocalProject(
                    git_url=upstream_url, working_dir=self.upstream_repo_path
                )
        else:
            # $CWD is the upstream repo and we just need to pick
            # downstream stuff
            self.upstream_repo_path = self.local_project.working_dir
            self.upstream_lp = self.local_project

    @property
    def primary_archive(self) -> Path:
        if not self._primary_archive:
            self._primary_archive = self.dist_git.download_upstream_archive()
        return self._primary_archive

    @property
    def dist_git(self) -> DistGit:
        if not self._dist_git:
            self._dist_git = self._get_dist_git()
            # we need to parse the spec twice
            # https://github.com/rebase-helper/rebase-helper/issues/848
            self._dist_git.download_remote_sources()
            self._dist_git.specfile.reload()
        return self._dist_git

    @property
    def upstream_ref(self) -> str:
        if self._upstream_ref is None:
            self._upstream_ref = get_tarball_comment(str(self.primary_archive))
            if self._upstream_ref:
                logger.info(
                    "upstream base ref was not set, "
                    f"discovered it from the archive: {self._upstream_ref}"
                )
            else:
                # fallback to HEAD
                try:
                    self._upstream_ref = self.local_project.commit_hexsha
                except ValueError as ex:
                    raise PackitException(
                        "Current branch seems to be empty - we cannot get the hash of "
                        "the top commit. We need to set upstream_ref in packit.yaml to "
                        "distinct between upstream and downstream changes. "
                        "Please set --upstream-ref or pull the upstream git history yourself. "
                        f"Error: {ex}"
                    )
                logger.info(
                    "upstream base ref was not set, "
                    f"falling back to the HEAD commit: {self._upstream_ref}"
                )
        return self._upstream_ref

    @property
    def specfile_path(self) -> Path:
        return self.dist_git.get_root_downstream_dir_for_source_git(
            self.local_project.working_dir
        ).joinpath(self.dist_git.absolute_specfile_path.name)

    def _get_dist_git(
        self,
    ) -> DistGit:
        """
        For given package names, clone the dist-git repo in the given directory
        and return the DistGit class

        :return: DistGit instance
        """
        if self.centos_package:
            self.dist_git_branch = self.dist_git_branch or "c9s"
            # let's be sure to cover anything 9 related,
            # even though "c9" will probably never be a thing
            if "c9" in self.dist_git_branch:
                return CentOS9DistGit.clone(
                    config=self.config,
                    package_config=self.package_config,
                    path=self.dist_git_path,
                    branch=self.dist_git_branch,
                )
            return CentOS8DistGit.clone(
                config=self.config,
                package_config=self.package_config,
                path=self.dist_git_path,
                branch=self.dist_git_branch,
            )
        else:
            # If self.dist_git_branch is None we will checkout/store repo's default branch
            dg = DistGit.clone(
                config=self.config,
                package_config=self.package_config,
                path=self.dist_git_path,
                branch=self.dist_git_branch,
            )
            self.dist_git_branch = (
                self.dist_git_branch or dg.local_project.git_project.default_branch
            )
            return dg

    def _pull_upstream_ref(self):
        """
        Pull the base ref from upstream to our source-git repo
        """
        # fetch operation is pretty intense
        # if upstream_ref is a commit, we need to fetch everything
        # if it's a tag or branch, we can only fetch that ref
        self.local_project.fetch(
            str(self.upstream_lp.working_dir), "+refs/heads/*:refs/remotes/origin/*"
        )
        self.local_project.fetch(
            str(self.upstream_lp.working_dir),
            "+refs/remotes/origin/*:refs/remotes/origin/*",
        )
        try:
            next(self.local_project.get_commits())
        except GitCommandError as ex:
            logger.debug(f"Can't get next commit: {ex}")
            # the repo is empty, rebase would fail
            self.local_project.reset(self.upstream_ref)
        else:
            self.local_project.rebase(self.upstream_ref)

    def _run_prep(self):
        """
        run `rpmbuild -bp` in the dist-git repo to get a git-repo
        in the %prep phase so we can pick the commits in the source-git repo
        """
        _packitpatch_path = shutil.which("_packitpatch")
        if not _packitpatch_path:
            raise PackitException(
                "We are trying to unpack a dist-git archive and lay patches on top "
                'by running `rpmbuild -bp` but we cannot find "_packitpatch" command on PATH: '
                "please install packit as an RPM."
            )
        logger.info(
            f"expanding %prep section in {self.dist_git.local_project.working_dir}"
        )

        rpmbuild_args = [
            "rpmbuild",
            "--nodeps",
            "--define",
            f"_topdir {str(self.dist_git.local_project.working_dir)}",
            "-bp",
            "--define",
            f"_specdir {str(self.dist_git.absolute_specfile_dir)}",
            "--define",
            f"_sourcedir {str(self.dist_git.absolute_source_dir)}",
        ]
        rpmbuild_args += RPM_MACROS_FOR_PREP
        if logger.level <= logging.DEBUG:  # -vv can be super-duper verbose
            rpmbuild_args.append("-v")
        rpmbuild_args.append(str(self.dist_git.absolute_specfile_path))

        run_command(
            rpmbuild_args,
            cwd=self.dist_git.local_project.working_dir,
            print_live=True,
        )

    def _put_downstream_sources(self, lookaside_sources: List[str]):
        """
        place sources from the downstream into the source-git repository
        """
        if self.dist_git_branch:
            self.dist_git.checkout_branch(self.dist_git_branch)
        self.dist_git.download_upstream_archive()
        root_downstream_dir = self.dist_git.get_root_downstream_dir_for_source_git(
            self.local_project.working_dir
        )
        os.makedirs(root_downstream_dir, exist_ok=True)

        shutil.copy2(self.dist_git.absolute_specfile_path, root_downstream_dir)

        logger.info(
            f"Copy all sources from {self.dist_git.absolute_source_dir} to {root_downstream_dir}."
        )

        # we may not want to copy the primary archive - it's worth a debate
        for source in self.dist_git.specfile.get_sources():
            if Path(source).name in lookaside_sources:
                logger.debug(
                    f"Source {source} will be fetched from the lookaside cache."
                )
                continue
            source_dest = root_downstream_dir / Path(source).name
            logger.debug(f"copying {source} to {source_dest}")
            shutil.copy2(source, source_dest)

        self.local_project.stage(self.dist_git.source_git_downstream_suffix)
        self.local_project.commit(message="add downstream distribution sources")

    def _get_lookasisde_sources(self) -> List[Dict[str, str]]:
        """
        Read "sources" file from the dist-git repo and return a list of dicts
        with path and url to sources stored in the lookaside cache
        """
        try:
            config = LookasideCacheHelper._read_config(self.config.fedpkg_exec)
            base_url = config["lookaside"]
        except (configparser.Error, KeyError) as e:
            raise LookasideCacheError("Failed to read rpkg configuration") from e

        package = self.dist_git.package_config.downstream_package_name
        basepath = self.dist_git.local_project.working_dir

        sources = list()
        for source in LookasideCacheHelper._read_sources(basepath):

            if self.config.fedpkg_exec == "fedpkg":
                url = "{0}/{1}/{2}/{3}/{4}/{2}".format(
                    base_url,
                    package,
                    source["filename"],
                    source["hashtype"],
                    source["hash"],
                )
            else:  # centpkg
                url = "{0}/{1}/{2}/{3}/{2}".format(
                    base_url, package, source["filename"], source["hash"]
                )

            path = source["filename"]
            sources.append({"path": path, "url": url})

        return sources

    def _add_packit_config(self, lookaside_sources: List[Dict[str, str]]):
        """
        Add .packit.yaml config to the sources-git repo
        """
        # mypy wanted this type annotation -_-
        default_packit_yaml: Dict[str, Union[str, List[str], List[Dict[str, str]]]] = {
            "specfile_path": f"{self.specfile_path.relative_to(self.local_project.working_dir)}",
            "upstream_ref": self.upstream_ref,
            "patch_generation_ignore_paths": [
                self.dist_git.source_git_downstream_suffix
            ],
        }

        if lookaside_sources:
            default_packit_yaml["sources"] = lookaside_sources

        if self.upstream_lp.git_url:
            default_packit_yaml["upstream_project_url"] = self.upstream_lp.git_url

        packit_yaml_path = self.local_project.working_dir.joinpath(".packit.yaml")
        packit_yaml_path.write_text(
            yaml.dump_all(
                [default_packit_yaml],
                Dumper=SafeDumperWithoutAliases,
                # default_flow_style=False dumps things into a block instead of ugly inline
                # inline example:
                #   key: {key1: value1, key2: value2}
                # block example:
                #   key:
                #     key1: value1
                #     key2: value1
                # True in el8, False in the latest pyyaml
                default_flow_style=False,
            )
        )

        self.local_project.stage(".packit.yaml")
        self.local_project.commit("add packit.yaml")

    def get_BUILD_dir(self):
        path = self.dist_git.local_project.working_dir
        build_dirs = [d for d in (path / "BUILD").iterdir() if d.is_dir()]
        if len(build_dirs) > 1:
            raise RuntimeError(f"More than one directory found in {path / 'BUILD'}")
        if len(build_dirs) < 1:
            raise RuntimeError(f"No subdirectory found in {path / 'BUILD'}")
        return build_dirs[0]

    def _rebase_patches(self, from_branch):
        """Rebase current branch against the from_branch"""
        to_branch = "dist-git-commits"  # temporary branch to store the dist-git history
        logger.info(f"Rebase patches from dist-git {from_branch}.")
        BUILD_dir = self.get_BUILD_dir()
        self.local_project.fetch(BUILD_dir, f"+{from_branch}:{to_branch}")

        # shorter format for better readability in case of an error
        commits_to_cherry_pick = [
            c.hexsha[:8]
            for c in LocalProject(working_dir=BUILD_dir).get_commits(from_branch)
        ][-2::-1]
        if commits_to_cherry_pick:
            self.local_project.git_repo.git.cherry_pick(
                *commits_to_cherry_pick,
                keep_redundant_commits=True,
                allow_empty=True,
                strategy_option="theirs",
            )
        self.local_project.git_repo.git.branch("-D", to_branch)

    def create_from_upstream(self):
        """
        create a source-git repo from upstream
        """
        self._pull_upstream_ref()
        lookaside_sources = self._get_lookasisde_sources()
        self._put_downstream_sources([di["path"] for di in lookaside_sources])
        self._add_packit_config(lookaside_sources)
        if self.dist_git.specfile.get_applied_patches():
            self._run_prep()
            self._rebase_patches(
                get_default_branch(
                    LocalProject(working_dir=self.get_BUILD_dir()).git_repo
                )
            )
            # TODO: patches which are defined but not applied should be copied
